/*
 * Component Name:
 *      agent_util.c
 *
 * $Date: 5/23/01 17:21:44 $
 *
 * $Source: src/eif/agent_util.c, tec, tec_3.7.1 $
 *
 * $Revision: 1.82 $
 *
 * Description:
 *      common functions for TME and non-TME agent libraries
 *
 * Note:
 *      Tab stops = 3
 *
 * (C) COPYRIGHT TIVOLI Systems, Inc. 1994
 * Unpublished Work
 * All Rights Reserved
 * Licensed Material - Property of TIVOLI Systems, Inc.
 */

#ifdef OS400
#   include "QymaCpRight.h"
#   include "qymadef.h"
#   include <micomput.h>
#   include <ifs.h>
#else
#   include <stdio.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>

#ifndef WIN32
#	include <sys/socket.h>
#else
#	include <winsock.h>
#	include "cpl/msvc_stat_macros.h"
#endif

#ifndef PC
#	include <sys/time.h>
#endif

#ifdef OS2
#	include <netinet\in.h>
#	include <netdb.h>
#	include <signal.h>
#endif

#include "slist.h"
#include "eif_common.h"
#include "eifdebug.h"
#include "tec_eif.h"
#include "regexp.h"

#ifndef NO_UTF8
	#if defined(ENDPOINT_BUILD) || defined(ADAPTERS_ONLY_UNIX) || defined(NW4) || defined(OS400)
		#include <mrt/tis.h>
	#else
		#include <tivoli/tis.h> 	
	#endif 
	#include "tec_delims.h"
#endif

#ifdef WIN32
#	define strdup _strdup
#	define stat  _stat
#endif

#if defined(OS400)    /* for V3R1, inline is not supported */
#pragma inline ( fatal_memory_error )
#pragma inline ( tec_env_delete )
#pragma inline ( tec_event_delete )
#endif

#if defined(NETWARE)
#	if !defined(ANCIENT_SDK)
	/*
	**	With more recent NetWare SDKs, stat() and fstat() are macro aliases
	**	to stat_411() and fstat_411().  Since these symbols don't exist for
	**	NetWare 4.10, undefine these macros to force referencing the real
	**	stat() and fstat() symbols.
	*/
#		if defined(stat)
#			undef stat
#		endif
#		if defined(fstat)
#			undef fstat
#		endif
		/*
		** With more recent NetWare SDKs, ctype functions are macro aliases that
		** reference the symbol __ctype which is not present in NetWare 4.10.
		** Undefine these macros and force referencing the actual ctype function
		** symbols.
		*/
#		if defined(isspace)
#			undef isspace
#		endif
#	endif /* !defined(ANCIENT_SDK) */

	/*
	**	Default configuration file path.
	*/

	/*
	**	For ConsolePrintf...
	*/
#	if defined(ANCIENT_SDK)
#		include <conio.h>
#	else
#		include <nwconio.h>
#	endif
#endif /* defined(NETWARE) */

#ifdef NETWARE
#	define DEFAULT_CONF_FILE "SYS:ETC\\TIVOLI\\TECADEIF.CFG"
#else
#	define DEFAULT_CONF_FILE "/etc/Tivoli/tecad_eif.conf"
#endif

#define RETRY_WINDOW        0

/*
 * Global Variables.
 * We need to move away from having any global variables.  This last one
 * provides backward compatibility.  Newer APIs (tec_create_handle_r())
 * provide for a dynamic AgentConfig and true support for multiple
 * simultaneous configurations/connections.
 */
static AgentConfig	default_agent_info = { {NULL}, NULL };

int TECAD_no_connection = FALSE; /* did the last tec_put_event() fail? */
int tec_errno = 0;

typedef struct {
	char *keyword;
	char *value;
} tec_env_t;

/*
 * Data structure used for storing slot value for filtering
 */

typedef struct {
	char   *name;
	char   *value;
   regexp *regex;
} ev_slot;

#define SUBKEY_SOURCE           "Source"
#define SUBKEY_EVENT_ID         "EventId"
#define MAX_EVENT_SRCS          16
#define MAX_EVENT_IDS           16
#ifdef WIN32
#define SUBKEY_GUI_EVENT_ID     "EventID"
#define SUBKEY_LOG		"Log"
#define SUBKEY_EVENT_TYPE	"EventType"
#define MAX_EVENT_LOGS           5
#define MAX_EVENT_TYPES          6
#elif defined(NETWARE)
#define SUBKEY_SEVERITY         "Severity"
#define SUBKEY_LOCUS            "Locus"
#define SUBKEY_CLASS            "Class"
#define MAX_EVENT_SEVERITIES    16
#define MAX_EVENT_LOCUSES       16
#define MAX_EVENT_CLASSES       16
#endif

typedef struct {
	int	event_id[MAX_EVENT_IDS + 1];
	char	*event_src[MAX_EVENT_SRCS + 1];
#ifdef WIN32
	char	*event_log[MAX_EVENT_LOGS + 1];
	char	*event_type[MAX_EVENT_TYPES + 1];
#elif defined(NETWARE)
	int	event_severity[MAX_EVENT_SEVERITIES + 1];
	int	event_locus[MAX_EVENT_LOCUSES + 1];
	int	event_class[MAX_EVENT_CLASSES + 1];
#endif
}  NT_pre_filter;

static void tec_env_delete(void *);
static int tec_env_search(void *, void *);
static int tec_add_filter(AgentConfig *, char *, int);

#ifdef OS400
   int startup_msgs = TRUE;
#endif

/*
 * Callback functions for filter list
 */
static int	tec_pre_filter_walk(void *, void *);
static int	tec_filter_walk(void *, void *);
static void	tec_filter_delete(void *);
static void	tec_event_delete(void *);
static void	free_prefilter(NT_pre_filter *);
static int	tec_add_prefilter(AgentConfig *, char *);
static void	tec_pre_filter_delete(void *);

#ifndef NO_UTF8
static void tec_utf8_init(tec_handle_common_t *);
#endif

/*
 * IY25233 - This will allow the timeout times in tec_pmap_getport to change.
 * These are their default values.
 */
time_t  GETPORT_TIMEOUT_SECONDS = 0;
long    GETPORT_TIMEOUT_USEC = 50000;
time_t  GETPORT_TOTAL_TIMEOUT_SECONDS = 0;
long    GETPORT_TOTAL_TIMEOUT_USEC = 50000;
int     EIPC_VERIFY_PORTMAPPER = 0;


void
fatal_memory_error(VOID_DECL)
{
#ifdef OS400
	sendPgmMsg(TEC_MEM_FAIL, TEC_ADP_MSG_FILE_Q, " ", 0,
		   "*ESCAPE   ");
#else
	fprintf(stderr, "Memory allocation failure, exiting...\n");
#endif
	exit(1);
}

int
tec_filter_event(tec_handle_t handle, char *event, int filter_type)
{
	return(common_filter_event(
      (tec_handle_common_t*)handle, event, filter_type));
}

int
tec_pre_filter_event(tec_handle_t handle, PreFilterEvent *event)
{
	return(common_pre_filter_event((tec_handle_common_t*)handle, event));
}

int
tec_buffer_event(tec_handle_t handle, char *event)
{
	return(common_buffer_event((tec_handle_common_t*)handle, event));
}

/*
 * Purpose:
 *    Read a configuration file line.
 *
 * Description:
 * 	Keeps reading into a buffer until a newline is encountered.
 * 	Splits line into keywork and value.
 * 	Uses internal, static buffer if the line fits within it.
 *
 * Returns:
 * 	Return value is NULL when the end of file is reached.
 * 	Keywork and value pointers are returned through the parameter list.
 *
 * Notes/Dependencies:
 */

static char *
read_conf_line(FILE *fp, char **keyword_ptr, char **value_ptr)
{
	char			*whole_line;
	char			*delimiter;
	int			read_len;
	int			end_of_line;
	static char	workbuf[WORKBUF_SIZE];

	*keyword_ptr = *value_ptr = NULL;

	/*
	 * Read until a newline is found.  Concatenate each read into one big line.
	 */
	for (whole_line = NULL, end_of_line = 0; !end_of_line; ) {
		if (fgets((char *) workbuf, WORKBUF_SIZE, fp) == NULL)
			return(NULL);
		else {  /* (fgets((char *) workbuf, WORKBUF_SIZE, fp) != NULL) */
			read_len = strlen(workbuf);

			if ((workbuf[read_len - 1] == '\n')
#ifdef OS400
			|| (workbuf[read_len - 1] == 0x25)
#endif
                        ) {
				workbuf[--read_len] = '\0';
#ifdef OS400
				workbuf[triml(workbuf, ' ')] = '\0';
#endif
				end_of_line = 1;
			}

			if (whole_line != NULL) {
				whole_line = (char *)realloc((void *) whole_line,
													(size_t)(strlen(whole_line) + read_len + 1));
				NULL_CHECK(whole_line);
				strcat(whole_line, workbuf);
			}
			else {   /* (whole_line == NULL) */
				whole_line = (char *) malloc((size_t) read_len + 1);
				NULL_CHECK(whole_line);
				strcpy(whole_line, workbuf);
			}   /* (whole_line != NULL) */
		}   /* (fgets((char *) workbuf, WORKBUF_SIZE, fp) == NULL)  */
	}   /* (whole_line = NULL, end_of_line = 0; !end_of_line; ) */

	/*
	 * Split the line into keyword and value
	 */
	if ((whole_line[0] != '#') && (whole_line[0] != '\n')) {

		/*
		 * Check if this is a filter specification
		 */
		if (strncmp(whole_line, FILTER, strlen(FILTER)) == 0) {
			delimiter = whole_line + strlen(FILTER) - 1;
			*keyword_ptr = whole_line;
			*value_ptr = delimiter + 1;
		}

		/*
		 * Check if this is a cache filter specification
		 */
		else if (strncmp(whole_line, FILTER_CACHE, strlen(FILTER_CACHE)) == 0) {
			delimiter = whole_line + strlen(FILTER_CACHE) - 1;
			*keyword_ptr = whole_line;
			*value_ptr = delimiter + 1;
		}

		/*
		 * Check if this is a pre_filter specification
		 */
		else if (strncmp(whole_line, PREFILTER, strlen(PREFILTER)) == 0) {
			delimiter = whole_line + strlen(PREFILTER) - 1;
			*keyword_ptr = whole_line;
			*value_ptr = delimiter + 1;
		}

		/*
		 * All other specifications
		 */
		else if ((delimiter = strchr(whole_line, '=')) != NULL) {
			*delimiter = '\0';
			*keyword_ptr = whole_line;
			*value_ptr = delimiter + 1;
		}
	}   /* ((whole_line[0] != '#') && (whole_line[0] != '\n')) */

	return(whole_line);
}

/*
 * Purpose:
 *    initialization for TEC agent
 *
 * Description:
 *
 * Returns:
 *
 * Notes/Dependencies:
 */

int
common_agent_init(char *cfgfile, AgentConfig **config_ptr)
{
	AgentConfig	*agent_info;
        tec_event_cache_t    evCacheTmp;
        tec_event_cache_t   *pEC = NULL;
	FILE			*fp;
	char			*conf_line;
	char			*keyword;
	char			*value;
	tec_env_t	*entry;
	struct stat	sbuf;
	char			cfg_file[FILE_LEN];
	int			line_no = 1;
#ifndef OS400
	char                    default_cfg[] = DEFAULT_CONF_FILE;
#else
	char                    replace_data[300];
	int                     pos;
	char                    default_cfg[] = TEC_CFG_MSGQ_IFS;
#endif

	/*
	 * If a config. info pointer is not NULL, allocate space and set the pointer.
	 * Otherwise, ues the global config. structure.
	 */
	if (config_ptr != NULL) {
		*config_ptr = (AgentConfig *) calloc(1, sizeof(AgentConfig));
		agent_info = *config_ptr;
		NULL_CHECK (agent_info);
	}
	else
		agent_info = &default_agent_info;

	/*
	 * Initialize the env list and the filter list
	 */
	if (agent_info->tec_env_list != NULL)
		SL_DelAll(agent_info->tec_env_list, tec_env_delete);

	if (agent_info->tec_pre_filters != NULL)
		SL_DelAll(agent_info->tec_pre_filters, tec_pre_filter_delete);

	if (agent_info->tec_event_filters != NULL)
		SL_DelAll(agent_info->tec_event_filters, tec_filter_delete);

	if (agent_info->tec_event_cache_filters != NULL)
		SL_DelAll(agent_info->tec_event_cache_filters, tec_filter_delete);

	if ((agent_info->tec_env_list = SL_Create()) == NULL)
		fatal_memory_error();

	/*
	 * Create structured list to store filters
	 */
	if ((agent_info->tec_pre_filters = SL_Create()) == NULL)
		fatal_memory_error();

	if ((agent_info->tec_event_filters = SL_Create()) == NULL)
		fatal_memory_error();

	if ((agent_info->tec_event_cache_filters = SL_Create()) == NULL)
		fatal_memory_error();

	if (cfgfile == NULL) {
		if (stat(default_cfg, &sbuf) == 0)
			strcpy(cfg_file, default_cfg);
		else
			return(0);
	}
	else
		strcpy(cfg_file, cfgfile);

	/*
	 * Check if configuration file exists & is a valid file
	 */
	if (stat(cfg_file, &sbuf) != 0) {
#ifndef OS400
		perror("stat failed");
#else
		sendPgmMsg(TEC_CANT_OPEN, TEC_ADP_MSG_FILE_Q, cfg_file, strlen(cfg_file),
			   "*DIAG     ");
#endif
		return(2);
	}

#ifndef OS2
	if (!S_ISREG(sbuf.st_mode)) {
#else
	if ((sbuf.st_mode & S_IFREG) == 0) {
#endif
#ifndef OS400
		printf("%s: Not a regular file\n", cfg_file);
#else
		sendPgmMsg(TEC_BAD_FILE, TEC_ADP_MSG_FILE_Q, cfg_file, strlen(cfg_file),
			   "*DIAG     ");
#endif
		return(3);
	}

	/*
	 * Open the given configuration file
	 */
#ifdef OS400
	if (!(fp = fopen(cfg_file, "r, ccsid=37"))) {
		sendPgmMsg(TEC_CANT_OPEN, TEC_ADP_MSG_FILE_Q, cfg_file, strlen(cfg_file),
			   "*DIAG     ");
#else
	if (!(fp = fopen(cfg_file, "r"))) {
		printf("%s: Can not fopen\n", cfg_file);
#endif
		return(4);
	}

	/*
	 * Process file lines
	 */
	while ((conf_line = read_conf_line(fp, &keyword, &value)) != NULL) {
		line_no++;

		/*
		 * Lines begining with '#' are comments
		 */
		if ((keyword != NULL) && (value != NULL)) {
			/*
			 * Check if this is a prefilter specification
			 */
			if (strncmp(keyword, PREFILTER, strlen(PREFILTER)) == 0) {
				if (tec_add_prefilter(agent_info, value) == -1) {
#ifndef OS400
					printf("syntax error in PreFilter, line %d ignored\n",line_no);
#else
					if ( startup_msgs )
					{
						sprintf(replace_data, "%04d%s", line_no, cfg_file);
						sendPgmMsg(TEC_SNTX_ERR, TEC_ADP_MSG_FILE_Q, replace_data,
							strlen(replace_data), "*INFO     " );
					}
#endif

				}
				free((void *) conf_line);
			}

			/*
			 * Check if this is a filter specification
			 */
			else if (strncmp(keyword, FILTER, strlen(FILTER)) == 0) {
				if (tec_add_filter(agent_info, value, FILTER_EVENT_TYPE) < 0) {
#ifndef OS400
					printf("syntax error in configuration file, line %d ignored\n",
																						 line_no);
#else
					if ( startup_msgs )
					{
						sprintf(replace_data, "%04d%s", line_no, cfg_file);
						sendPgmMsg(TEC_SNTX_ERR, TEC_ADP_MSG_FILE_Q, replace_data,
							strlen(replace_data), "*INFO     " );
					}
#endif
				}
				free((void *) conf_line);
			}


			/*
			 * Check if this is a cache filter specification
			 */
			else if (strncmp(keyword, FILTER_CACHE, strlen(FILTER_CACHE)) == 0) {
				if (tec_add_filter(agent_info, value,FILTER_EVENT_CACHE_TYPE) < 0) {
#ifndef OS400
					printf("syntax error in configuration file, line %d ignored\n",
																						 line_no);
#else
					if ( startup_msgs )
					{
						sprintf(replace_data, "%04d%s", line_no, cfg_file);
						sendPgmMsg(TEC_SNTX_ERR, TEC_ADP_MSG_FILE_Q, replace_data,
							strlen(replace_data), "*INFO     " );
					}
#endif
				}
				free((void *) conf_line);
			}
			/* 
			 * IY25233 - Read in the timeout times from the "*.conf" file.
			 */
			else if (strncmp(keyword, "getport_timeout_seconds", strlen("getport_timeout_seconds")) == 0) {
				GETPORT_TIMEOUT_SECONDS=(time_t) atoi(value);
			}

			else if (strncmp(keyword, "getport_timeout_usec", strlen("getport_timeout_usec")) == 0) {
				GETPORT_TIMEOUT_USEC=(long) atoi(value);
			}
			
			else if (strncmp(keyword, "getport_total_timeout_seconds", strlen("getport_total_timeout_seconds")) == 0) {
				GETPORT_TOTAL_TIMEOUT_SECONDS=(time_t) atoi(value);
			}
			
			else if (strncmp(keyword, "getport_total_timeout_usec", strlen("getport_total_timeout_usec")) == 0) {
				GETPORT_TOTAL_TIMEOUT_USEC=(long) atoi(value);
			}
            else if (strncmp(keyword, "eipc_verify_portmapper", strlen("eipc_verify_portmapper")) == 0) {
                if (strcmp(value, "YES") == 0)
                    EIPC_VERIFY_PORTMAPPER = 1;
            }
			else {

				/*
				 * Create value list entry
				 */
				if ((entry = (tec_env_t *) malloc(sizeof(tec_env_t))) == NULL)
					fatal_memory_error();

				/*
				 * Create entry keyword
				 */
				entry->keyword = strdup(keyword);
				entry->value = strdup(value);
            if (entry->keyword == NULL || entry->value == NULL)
               fatal_memory_error();

				/*
				 * Insert entry in list
				 */
				if (SL_InsLast(agent_info->tec_env_list, (void *) entry) == FALSE)
					fatal_memory_error();

            free(conf_line);
			}
		}
		else if (strlen(conf_line)
					&& (conf_line[0] != '#') && (conf_line[0] != '\n')
#ifdef OS400
					&& (conf_line[0] != 0x15)
#endif
		) {
#ifndef OS400
			printf("syntax error in configuration file, line %d ignored\n",line_no);
#else
			if ( startup_msgs )
			{
				sprintf(replace_data, "%04d%s", line_no, cfg_file);
				sendPgmMsg(TEC_SNTX_ERR, TEC_ADP_MSG_FILE_Q, replace_data,
					   strlen(replace_data), "*INFO     " );
			}
#endif
			free((void *) conf_line);
		}
		else
			free((void *) conf_line);
	}   /* while ((conf_line = read_conf_line(fp, &keyword, &value)) != NULL) */

	/*
	 * Close the configuration file.
	 */
	fclose(fp);

	/*
	 * Initialize some flags.
	 */
	agent_info->filter_in = agent_match(agent_info, TECAD_FILTERMODE, "IN");
	agent_info->pre_filter_in = agent_match(agent_info, TECAD_PFLTMODE, "IN");

	/*
	 * Initialize buffer configuration.
	 */
	tec_buffer_init(agent_info);
        if (agent_info->buffer_info.do_buffer)
        {
                pEC = &evCacheTmp;
                cache_constructor(pEC);
                if (!cache_open_file(pEC, agent_info->buffer_info.buffer_path, agent_info->buffer_info.max_size))
                {
#ifndef OS400
                        printf("error in opening cache file %s,\n",agent_info->buffer_info.buffer_path);
#else
                        sendPgmMsg(TEC_CANT_OPEN, TEC_ADP_MSG_FILE_Q, agent_info->buffer_info.buffer_path, strlen(agent_info->buffer_info.buffer_path),
                                   "*DIAG     ");
#endif
                         return(5);
                }
                else
                        cache_destructor(pEC);
        }
	return(0);
}

/*
 * Purpose:
 *    Initialization for TEC common fields in an EIF handle.
 *
 * Description:
 *
 * Returns:
 * 	Success - 0
 * 	Failure - Positive error number
 *
 * Notes/Dependencies:
 */

int
common_agent_handle_init(tec_handle_common_t	*common_h, char *conf_file,
								 char *location, char *default_loc)
{
	char		**hosts;
	char		*env;
	int		j;

	/*
	 * If a configuration file name was specified, then initialize the agent
	 * configuration structure from it.
	 * Otherwise, assume that tec_agent_init() was called earlier and the
	 * configuration is in the global default_agent_config variable.
	 */
	if (conf_file != NULL) {
		j = common_agent_init(conf_file, &common_h->agent_config);

		if (j > 0)
			return(j);
	}
	else
		common_h->agent_config = &default_agent_info;

	common_h->dont_filter = 0;
	common_h->test_mode = FALSE;
	common_h->test_out = NULL;
	common_h->max_locations = 0;
	common_h->location[0] = NULL;
	common_h->in_flush_events = 0;
	common_h->retry_window = 0;
	common_h->no_utf8_conversion=0;


   /* Initialize event cache struct. */
   cache_constructor(&common_h->event_cache);

#ifndef NO_UTF8
	/* initialize data structures for utf8 conversions */
	tec_utf8_init(common_h);
#endif
	/*
	 * Initialize server location(s)
	 */

	tecadINITIAL(DEBUGFILE_AGENT_COMM)
	tecadDEBUG_T(DEBUGFILE_AGENT_COMM, "entering common_agent_handle_init")

	if (location == NULL) {
		tecadDEBUG_T(DEBUGFILE_AGENT_COMM, "location is NULL")
		hosts = common_split_entries(common_h, "ServerLocation",
											  &common_h->max_locations);

		if (common_h->max_locations == 0) {
			if (default_loc != NULL) {
				common_h->max_locations = 1;
				common_h->location[0] = strdup(default_loc);
				NULL_CHECK(common_h->location[0]);
			}
			else {
				/* return(E_HOST); */
				common_h->max_locations = 1;
#if !defined(NETWARE)
				common_h->location[0] = NULL;
#else  /* defined(NETWARE) */
		        common_h->location[0] = strdup("127.0.0.1");
				NULL_CHECK(common_h->location[0]);
		        printf("Please set the \"ServerLocation\" parameter and restart the adapter\n");
		        ConsolePrintf("Please set the \"ServerLocation\" parameter and restart the adapter\r\n");
#endif /* defined(NETWARE) */

			}
		}
		else{

	    /* rsm: we need to make the hosts string into multiple strings so we
	            can move 'em around later */
		    for (j = 0; j < common_h->max_locations; j++) {
		        common_h->location[j] = strdup(hosts[j]);
			NULL_CHECK(common_h->location[j]);
	            }
		}

      free(hosts);
      hosts = NULL;
	}
	else {
		int 	i;

		tecadDEBUG_T(DEBUGFILE_AGENT_COMM, "location is NOT NULL")
		common_h->location[0] = strdup(location);
		NULL_CHECK(common_h->location[0]);

		tecadDEBUG_S(DEBUGFILE_AGENT_COMM, common_h->location[0])

		hosts = common_split_entries(common_h, "ServerLocation",
											  &common_h->max_locations);

		for (i = 1;
			  (i < ((TECEIF_MAX_SERVERS > (common_h->max_locations + 1))
					  ? (common_h->max_locations + 1)
					  : TECEIF_MAX_SERVERS));
			  i++)
		{
			common_h->location[i]=strdup(hosts[i-1]);
			NULL_CHECK(common_h->location[i]);
			tecadDEBUG_S(DEBUGFILE_AGENT_COMM, common_h->location[i])
		}

		if (common_h->max_locations < TECEIF_MAX_SERVERS)
         common_h->max_locations++;

      free(hosts);
      hosts = NULL;
	}

	/*
	 * Check for test mode
	 */
	if (common_agent_match(common_h, "TestMode", "YES")) {
		common_h->test_mode = TRUE;
#ifdef OS400
		/* Check for library name on the front of the ServerLocation
		 * variable.  If not there, add "QUSRSYS", otherwise it'll
		 * open the file in the user's session's QTEMP location, which
		 * gets whacked when the user logs out.
		 */
		if(!strstr(common_h->location[0], "/"))
		{
		    char *temp_string = NULL,
		         add_string[] = "QUSRSYS/";

		    /* make room for the addition to the location string
		     * we already have
		     */
		    temp_string = (char *) malloc(sizeof(char) *
		      (strlen(common_h->location[0]) + strlen(add_string) + 1));
			NULL_CHECK(temp_string);
		    strcpy(temp_string, add_string);
		    strcat(temp_string, common_h->location[0]);
		    free(common_h->location[0]);
		    common_h->location[0] = temp_string;
		}
#endif

		if ((common_h->test_out = fopen(common_h->location[0], "w")) == NULL)
			return(errno);
	}


        /*
	 * Check for utf8 conversion
	 */
	if (common_agent_match(common_h, "NO_UTF8_CONVERSION", "YES")) {
		common_h->no_utf8_conversion = TRUE;
	}

	/*
	 * Look up the RetryWindow for (re-)connecting to servers.
	 */
	if ((env = agent_getenv(common_h->agent_config, "RetryInterval")) != NULL)
		common_h->retry_window = atoi(env);

	if (!common_h->retry_window)
		common_h->retry_window = RETRY_WINDOW;

	return(0);
}

int
common_agent_get_retry(tec_handle_common_t *handle)
{
	return(handle->retry_window);
}

/* ------------------------------------------------------------------------- */

static void
tec_env_delete(void *elem)
{
	tec_env_t *item = (tec_env_t *) elem;

	if (item != NULL) {
		if (item->keyword != NULL)
			free(item->keyword);

		free(item);
	}
}

/* ------------------------------------------------------------------------- */

char *
agent_getenv(AgentConfig *agent_info, char *keyword)
{
	tec_env_t				*found = NULL;

	if (keyword != NULL) {
		if (agent_info == NULL)
			agent_info = &default_agent_info;

		found = (tec_env_t *) SL_Find(agent_info->tec_env_list, keyword,
											tec_env_search);
	}
	return((found != NULL) ? found->value : NULL);
}

char *
tec_agent_getenv(char *keyword)
{
	return(tec_agent_getenv_r(NULL, keyword));
}

char *
tec_agent_getenv_r(tec_handle_t handle, char *keyword)
{
	tec_handle_common_t *th = (tec_handle_common_t *) handle;
	AgentConfig				*a_info;

	if (th != NULL)
 	   a_info = (th->agent_config) ? th->agent_config : NULL;
	else
	   a_info = NULL;

	return(agent_getenv(a_info, keyword));
}

static void free_tec_env_t(void *elem)
{
   tec_env_t *e = (tec_env_t*)elem;
   free(e->keyword);
   free(e->value);
   free(e);
}

/* tec_agent_setenv_r()
 *    Modifies <handle> so that the next time you call tec_agent_getenv_r()
 *    with <handle> and <keyword> it returns <value>.  You can pass <value>
 *    as NULL to erase a setting.  If you pass <handle> as NULL, then it
 *    modifies the global configuration initialized w/ tec_agent_init().
 *    Note: <keyword> and <value> are copied.
 */
void tec_agent_setenv_r(tec_handle_common_t *handle,
                        char *keyword,
                        char *value)
{
   AgentConfig *pAC;
   tec_env_t *oldEntry;

   if (handle == NULL)
      pAC = &default_agent_info;
   else
      pAC = handle->agent_config;

   /* Delete old entry if any. */
   oldEntry = (tec_env_t*)SL_Find(pAC->tec_env_list, keyword, tec_env_search);
   if (oldEntry != NULL)
      SL_Delete(pAC->tec_env_list, oldEntry, free_tec_env_t);

   /* Add new entry if value isn't NULL. */
   if (value != NULL)
   {
      tec_env_t *entry = malloc(sizeof(tec_env_t));
      if (entry == NULL)
         fatal_memory_error();
      entry->keyword = strdup(keyword);
      entry->value = strdup(value);
      if (entry->keyword == NULL || entry->value == NULL)
         fatal_memory_error();
   
      if (!SL_InsLast(pAC->tec_env_list, (void*)entry))
         fatal_memory_error();
   }
}

/* ------------------- TEC GATEWAY WORKAROUNDS ----------------------------- */

/*  Need to be able to turn buffering off in default config for
 *  the purposes of the TEC gateway.
 */
void gw_hack_turn_buffering_off()
{
   default_agent_info.buffer_info.do_buffer = FALSE;
}

/*  Need to be able to make the EIF believe events are being
 *  flushed on a handle when they really aren't for the purposes of the
 *   TEC gateway.
 */
void gw_hack_set_in_flush_events(tec_handle_common_t *h)
{
   h->in_flush_events = 1;
}

/*  Need to know whether we're running in the gateway or not at some
 *  places in the EIF for the purposes of the TEC gateway.
 */
int gw_hack_running_in_gateway = 0;

/* ------------------------------------------------------------------------- */

static int
tec_env_search(void *item_elem, void *keyword_elem)
{
	tec_env_t	*item = (tec_env_t *) item_elem;
	char			*keyword = (char *) keyword_elem;

	if ((item != NULL) && (item->keyword != NULL) && (keyword != NULL)) {
		if (strcmp(keyword, item->keyword))
			return (FALSE);
		else
			return (TRUE);
	}
	else
		return (FALSE);
}

/*
 * Syntax of Filter/FilterCache Specifications:
 * -------------------------------------------
 *   
 *   ws    := A sequence (possibly empty) characters satisfying the isspace()
 *            function from the C library.
 *
 *   value := A sequence (including the empty string) of characters excluding
 *            the equals sign and semi-colon.  The sequence of characters can
 *            contain whitespace in the middle, but not at the beginning or
 *            end.
 * 
 *   string := Begins and ends w/ a single quote character.  Any embedded single
 *             quotes must be escaped by a single quote.  The only other 
 *             restriction is that the string can't contain ASCII nul, carriage
 *             return, or linefeed.  The convention of escaping single quotes
 *             w/ single quotes derives from the BAROC instance parsing code
 *             in the rule engine.
 * 
 *   name   := <value> | <string>
 * 
 *   regexp := re:'<regular expression>'
 *             Embedded single quotes in the regular expression are escaped
 *             with a preceding '\' char since this is the convention for 
 *             escaping other special characters in the regular expression.
 *             Consult the code in the subdirectory 'regexp' for the 
 *             syntax of the regular expressions.
 * 
 *   pattern := <value> | <string> | <regexp>
 * 
 *   attrib_spec := <name>=<pattern>
 * 
 *   attrib_list := <attrib_spec> | <attrib_spec><ws>;<ws><attrib_list>
 * 
 *   attrib_specs := <empty_string> | ;<ws><attrib_list>
 * 
 *   filter_keyword := Filter | FilterCache
 * 
 *   filter := <filter_keyword>:<ws>Class<ws>=<ws><pattern><ws><attrib_specs>
 *
 * Examples:
 * --------
 *   Filters:
 *   1. Filter: Class=TEC_DB ; msg = I don't know
 *   2. Filter: Class=TEC_DB ; msg = I don''t know
 *   3. Filter:Class = TEC_DB ; msg = ' I don''t know'
 *   4. Filter: Class= re:'TEC_.*'
 *
 *   Events:
 *   1. TEC_DB;severity=CRITICAL;msg='I don''t know';END
 *   2. TEC_DB;severity=HARMLESS;msg=' I don''t know';END
 *   3. TEC_Stop;severity=WARNING;END
 *
 *   Filter 1 matches event 1.
 *   Filter 2 doesn't match any of the events.
 *   Filter 3 matches event 2.
 *   Filter 4 matches events 1 through 3.
 * 
 * Note that for filter 1 the single quote in the message value isn't escaped
 * because the value isn't enclosed in single quotes.  If you put two single
 * quotes in a row like I did in filter 2, that ends up meaning literally two
 * single quotes.  In filter 3 the message value is enclosed in single quotes
 * so the single quote escaping mechanism kicks in.
 *
 * BUG:
 * ---
 *   The value of a slot in a filter or an event can actually be a list
 *   such as
 *       Filter:Class=TEC_DB;acl=[admin,bob]
 *   In the above syntax definition a list is just interpreted as a <value>
 *   and extraneous whitespace is not stripped out as it should be.  For
 *   example, the event
 *       TEC_DB;acl=[admin, bob];END
 *   will NOT match the above filter because '[admin,bob]' is not exactly
 *   the same sequence of characters as '[admin, bob]' although they
 *   represent the same list.
 */

#define STATIC_TOKBUF_SIZE 512

typedef enum 
{
   EQUALS_TOK,
   SEMICOLON_TOK,
   STRING_TOK,
   REGEXP_TOK,
   EOF_TOK
} FilterTokType;

typedef struct
{
   char *token;
   int   parseErr;
   const char *data;
   int   bufsz;
   char *buf;
   char  staticBuf[STATIC_TOKBUF_SIZE];
} LexerState;

static void initLexerState(LexerState *ls, const char *s)
{
   ls->token = NULL;
   ls->data = s;
   ls->bufsz = STATIC_TOKBUF_SIZE;
   ls->buf = ls->staticBuf;
}

static void cleanupLexerState(LexerState *ls)
{
   if (ls->buf != ls->staticBuf)
   {
      free(ls->buf);
      ls->buf = NULL;
   }
}

static FilterTokType getNextToken(LexerState *ls, int findREs)
{
   int state = 0;
   const char *s = NULL;
   const char *e = NULL;
   const char *p = ls->data;
   char c;
   
   for (;;)
   {
      c = *p;
      switch (state)
      {
      case 0:  /* skip whitespace */
         if      (isspace(c))
            state = 0;
         else if (c == '\0')
            state = 15;
         else if (c == 'r')
         {
            s = e = p;
            state = 4;
         }
         else if (c == '\'')
         {
            s = e = p;
            state = 2;
         }
         else if (c == ';' || c == '=')
         {
            s = e = p;
            state = 9;
         }
         else
         {
            s = e = p;
            state = 1;
         }
         break;
      case 1:  /* scanning value tok */
         if      (c == '\0' || c == ';' || c == '=')
            state = 10;
         else if (isspace(c))
            state = 1;
         else
         {
            e = p;
            state = 1;
         }
         break;
      case 2:  /* scanning string tok */
         if      (c == '\'')
         {
            e = p;
            state = 3;
         }
         else if (c == '\0')
            state = 12;
         break;
      case 3:  /* saw single quote in string */
         if      (c == '\'')
         {
            e = p;
            state = 2;
         }
         else
            state = 11;
         break;
      case 4:  /* possible regexp - saw 'r', looking for 'e' */
         if      (c == '\0' || c == ';' || c == '=')
            state = 10;
         else if (c != 'e')
         {
            e = p;
            state = 1;
         }
         else
         {
            e = p;
            state = 5;
         }
         break;
      case 5:  /* possible regexp - saw 're', looking for ':' */
         if      (c == '\0' || c == ';' || c == '=')
            state = 10;
         else if (c != ':')
         {
            e = p;
            state = 1;
         }
         else
         {
            e = p;
            state = 6;
         }
         break;
      case 6:  /* possible regexp - saw 're:', looking for a single quote */
         if      (c == '\0' || c == ';' || c == '=')
            state = 10;
         else if (c != '\'')
         {
            e = p;
            state = 1;
         }
         else
         {
            e = p;
            state = 7;
         }
         break;
      case 7:  /* scanning regexp */
         if      (c == '\\')
         {
            e = p;
            state = 8;
         }
         else if (c == '\'')
         {
            e = p;
            state = 13;
         }
         else if (c == '\0')
            state = 14;
         break;
      case 8:  /* saw '\' in regexp, escapes next character. */
         if      (c == '\0')
            state = 14;
         else
            state = 7;
         break;
      default:
         assert(FALSE);
      }
      if (state <= 8)
         ++p;
      else
         break;
   }
   if (e == p)
      ++p;
   ls->data = p;

   /* If the token is too big to put in the current buffer then allocate a
    * bigger one. */
   if (e - s + 2 > ls->bufsz)
   {
      if (ls->buf != ls->staticBuf)
         free(ls->buf);
      ls->buf = (char*)malloc((int)(e - s + 2));
      if (ls->buf == NULL)
         fatal_memory_error();
      ls->bufsz =(int)( e - s + 2);
   }

   /* Unterminated regex or string? */
   ls->parseErr = (state == 12 || state == 14);

   ls->token = ls->buf;
   switch (state)
   {
   case 9:  /* semi-colon or equals */
      ls->buf[0] = *s;
      ls->buf[1] = '\0';
      return (*s == ';' ? SEMICOLON_TOK : EQUALS_TOK);
   case 10: /* value token */
   case 12: /* unterminated string, interpret as value token */
   case 14: /* unterminated regex, interpret as value token */
      memcpy(ls->buf, s, e - s + 1);
      ls->buf[e - s + 1] = '\0';
      return STRING_TOK;
   case 11: /* string token */
      {
         char *q;
         ++s;
         for (q = ls->buf; s < e; ++s, ++q)
            if (*s == '\'') *q = *s++;
            else            *q = *s;
         *q = '\0';
      }
      return STRING_TOK;
   case 13: /* regular expression */
      if (findREs)
      {
         s += 4;
         --e;
      }
      memcpy(ls->buf, s, e - s + 1);
      ls->buf[e - s + 1] = '\0';
      return (findREs ? REGEXP_TOK : STRING_TOK);
   case 15: /* EOF */
      ls->buf[0] = '\0';
      return EOF_TOK;
   default:
      assert(FALSE);
      /* suppress no return value warning. */
      return -1;
   }
}

static ev_slot* parseFilterOrEvent(const char *s, int isFilter, int *parseErr)
{
   LexerState ls;
   int sz = 4;
   int cnt = 0;
   char *name = NULL;
   int state = 0;
   FilterTokType tokt;
   int pe = FALSE; /* Parse Error */
   ev_slot *rv = (ev_slot*)calloc(sz,sizeof(ev_slot));
   int             utf8_msg_len = 0;
   unsigned char   *utf8_msg_buf = NULL;


   if (rv == NULL)
      fatal_memory_error();

   initLexerState(&ls, s);

   if (isFilter)
   {
      /* Skip "Class=" at the front of a filter */
      getNextToken(&ls, isFilter);
      pe |= (ls.parseErr || strcmp(ls.token, "Class") != 0);
      tokt = getNextToken(&ls, isFilter);
      pe |= (tokt != EQUALS_TOK);
   }

   name = (char*)malloc(6);
   if (name == NULL)
      fatal_memory_error();
   strcpy(name, "Class");
   state = 2;
   while ((tokt = getNextToken(&ls, isFilter)) != EOF_TOK)
   {
      switch (state)
      {
      case 0: /* looking for attribute name */
         pe |= (tokt != STRING_TOK);
         if (name != NULL)
             free(name);
         name = (char*)malloc(strlen(ls.token) + 1);
         if (name == NULL)
            fatal_memory_error();
         strcpy(name, ls.token);
         break;
      case 1: /* looking for equals sign */
         pe |= (tokt != EQUALS_TOK);
         break;
      case 2: /* looking for attribute value */
         if (cnt + 1 == sz)
         {
            /* Ran out of space, resize our buffer. */
            sz *= 2;
            rv = (ev_slot*)realloc(rv, sz*sizeof(ev_slot));
            if (rv == NULL)
               fatal_memory_error();
         }
         rv[cnt].name = name; name = NULL;
         rv[cnt].value = NULL;
         rv[cnt].regex = NULL;
         if (tokt == REGEXP_TOK)
         {
#ifdef NO_UTF8
            rv[cnt].regex = eif_regcomp(ls.token);
#else
#ifdef OS400
				/* IY26027 - For AS/400, don't convert to utf8 because the eif_regcomp function 
				 * looks for '*' which is different in ebcdic utf8.
				 */
            rv[cnt].regex = eif_regcomp(ls.token);
#else
    	    utf8_msg_len = strlen(ls.token) * TIS_TO_UTF8_RATIO + 1;
            utf8_msg_buf = (unsigned char *)malloc(utf8_msg_len);
            NULL_CHECK(utf8_msg_buf);
            memset(utf8_msg_buf, '\0', utf8_msg_len);
            tis_to_utf8(NULL, (unsigned char *) ls.token, -1, utf8_msg_buf, utf8_msg_len);
            rv[cnt].regex = eif_regcomp(utf8_msg_buf);
            free(utf8_msg_buf);
#endif
#endif
            pe |= (rv[cnt].regex == NULL);
         }
         if (rv[cnt].regex == NULL)
         {
            rv[cnt].value = (char*)malloc(strlen(ls.token) + 1);
            if (rv[cnt].value == NULL)
               fatal_memory_error();
            strcpy(rv[cnt].value, ls.token);
         }
         ++cnt;
         break;
      case 3: /* looking for semi-colon */
         pe |= (tokt != SEMICOLON_TOK);
         break;
      }
      ++state;
      state %= 4;
   }
   if (name != NULL)
   {
      free(name);
      name = NULL;
   }
   rv[cnt].name = NULL;
   rv[cnt].value = NULL;
   rv[cnt].regex = NULL;

   cleanupLexerState(&ls);
   *parseErr = pe;
   return rv;
}

static void deleteFilterOrEvent(ev_slot *filter)
{
   ev_slot *p = filter;
   for (; p->name != NULL; ++p)
   {
      free(p->name);
      if (p->value != NULL) free(p->value);
      if (p->regex != NULL) free(p->regex);
   }
   free(filter);
}

/*
 * Purpose:
 *    add an event filter to the list of filters
 *
 * Description:
 *
 * Returns:
 *    -1 on parse error
 *
 * Notes/Dependencies:
 */

static int
tec_add_filter(AgentConfig *agent_info, char *line, int filter_type)
{
   ev_slot *filter, *savefilter;
   int      parseErr;
   int             utf8_msg_len = 0;
   unsigned char   *utf8_msg_buf = NULL;


   filter = parseFilterOrEvent(line, TRUE, &parseErr);
   if (!parseErr && filter[0].name != NULL)
   {
#ifndef NO_UTF8
   /* convert slot values to UTF-8 before adding to the list of filters */ 
    savefilter= (ev_slot *) filter;

    for (; filter->name != NULL; ++filter)
    {
      if (filter->regex != NULL)
       continue;
      else
      {
       utf8_msg_len = strlen(filter->value) * TIS_TO_UTF8_RATIO + 1;
       utf8_msg_buf = (unsigned char *)malloc(utf8_msg_len);
       NULL_CHECK(utf8_msg_buf);
       memset(utf8_msg_buf, '\0', utf8_msg_len);
       tis_to_utf8(NULL, (unsigned char *) filter->value, -1, utf8_msg_buf, utf8_msg_len);

       /* setting filter->value to utf8_msg_buf does not work */
       free(filter->value);
       filter->value = (char *)malloc(utf8_msg_len);
       memset(filter->value, '\0', utf8_msg_len );
       strcpy(filter->value, utf8_msg_buf);
       free(utf8_msg_buf);
      }
    }

    filter = (ev_slot *) savefilter;

#endif

		switch (filter_type) {
      case FILTER_EVENT_CACHE_TYPE:
         if (SL_InsLast(agent_info->tec_event_cache_filters, (void *)filter) == FALSE)
            fatal_memory_error();
			break;

      case FILTER_EVENT_TYPE:
         if (SL_InsLast(agent_info->tec_event_filters, (void *)filter) == FALSE)
            fatal_memory_error();
			break;

      default:
         assert(FALSE);
		}
   }
   else
      deleteFilterOrEvent(filter);
   return (parseErr ? -1 : 0);
}

char *
tec_normalize_msg(char *in_message, unsigned long msg_len)
{
	char	*out_msg, *s, *suffix;

	if (!msg_len)
		return(NULL);

	/* AS/400 messages have the terminator string added in the SendEvent
         * function. Since the events are translated to UTF8 before coming 
         * into this function, this function would not be able to find the 
         * TECAD_EVENT_DELIMITER characters because AS/400 is EBCDIC. And 
         * EBCDIC translated to UTF8 is unreadable here.
         */
#ifdef OS400
	return(NULL);
#endif
	/*
	 * If it is already OK, just return
	 */
	if (msg_len >= TECAD_EVENT_DELIMITER_LEN) {
		if (!strncmp(&in_message[msg_len - TECAD_EVENT_DELIMITER_LEN],
						 TECAD_EVENT_DELIMITER,
						 (msg_len - TECAD_EVENT_DELIMITER_LEN)))
		{
			return(NULL);
		}
	}

	/*
	 * Have to allocate new space because we can't assume the current message is
	 * in a buffer that has room to grow.
	 */
	out_msg = malloc((size_t) (msg_len + TECAD_EVENT_DELIMITER_LEN + 1));
	NULL_CHECK(out_msg);
	strcpy(out_msg, in_message);

	if (msg_len < TECAD_EVENT_DELIMITER_LEN)
		suffix = out_msg;
	else
		suffix = &out_msg[msg_len - TECAD_EVENT_DELIMITER_LEN];

	if ((s = strstr(suffix,"END")) != NULL)
		strcpy(s, TECAD_EVENT_DELIMITER);
	else
		strcat(suffix, TECAD_EVENT_DELIMITER);

	return(out_msg);
}

/*
 * Purpose:
 *    Determine if an event should be filtered out, based upon the currently
 *		in-effect filters
 *
 * Description:
 *    The event must match ALL name-value pairs for at least one of the
 *		specified filters
 *
 * Returns:
 *    1 -> discard the event
 *    0 -> send the event
 *
 * Notes/Dependencies:
 *    If the input event string doesn't match the expected format, then
 *    tec_filter_event always returns 0.
 */

int
common_filter_event(tec_handle_common_t *ch, char *message, int filter_type)
{
	AgentConfig  *agent_info;
   ev_slot      *event, *saveevent;
	int           no_match;
   int           ignore;
   int           ret;
   int           utf8_msg_len = 0;
   unsigned char *utf8_msg_buf = NULL;

	agent_info = 
      (ch && ch->agent_config) ? ch->agent_config : &default_agent_info;

	/*
	 * If filtering IN,  then no matching filter means to discard the event.
	 * If filtering OUT, then no matching filter means to send    the event.
	 */
	no_match = (agent_info->filter_in) ? FILTER_DISCARD : FILTER_KEEP;

	/*
	 * If the appropriate filters are not defined, get out of here now.
	 * The list may exist but be empty, so check by trying to get the 1st elem.
	 */
	switch (filter_type) {
   case FILTER_EVENT_CACHE_TYPE:
      if (SL_GetPos(agent_info->tec_event_cache_filters, 0) == NULL) {
         return(no_match);
      }
      break;

   case FILTER_EVENT_TYPE:
      if (SL_GetPos(agent_info->tec_event_filters, 0) == NULL)
         return(no_match);
      break;

   default:
      assert(FALSE);
	}

   event = parseFilterOrEvent(message, FALSE, &ignore);

   /* IY24129 - convert each of the event slot values to UTF-8 before comparing
    * the values to the filter values (which are already converted to UTF8).
    */ 
    saveevent= (ev_slot *) event;

    for (; event->name != NULL; ++event)
    {
      if (event->regex != NULL)
       continue;
      else
      {
       utf8_msg_len = strlen(event->value) * TIS_TO_UTF8_RATIO + 1;
       utf8_msg_buf = (unsigned char *)malloc(utf8_msg_len);
       NULL_CHECK(utf8_msg_buf);
       memset(utf8_msg_buf, '\0', utf8_msg_len);
       tis_to_utf8(NULL, (unsigned char *) event->value, -1, utf8_msg_buf, utf8_msg_len);

       /* setting event->value to utf8_msg_buf does not work */
       free(event->value);
       event->value = (char *)calloc(utf8_msg_len, sizeof(char));
       strcpy(event->value, utf8_msg_buf);
       free(utf8_msg_buf);
      }
    }

    event = (ev_slot *) saveevent;

	switch (filter_type) {
   case FILTER_EVENT_CACHE_TYPE:
      ret = SL_Walk2(
         agent_info->tec_event_cache_filters, tec_filter_walk, event);
      break;

   case FILTER_EVENT_TYPE:
      ret = SL_Walk2(agent_info->tec_event_filters, tec_filter_walk, event);
      break;

   default:
      assert(FALSE);
	}

	if (agent_info->filter_in == 1)
		ret = !ret;

    deleteFilterOrEvent(event);
	return(ret);
}



/*
 * Purpose:
 *    match an event against a single filter from the list of possible filters
 *
 * Description:
 *    if the event matches ALL name-value pairs of the filter, that is a
 *  successful match
 *
 * Returns:
 *    0 -> successful match, the event will be discarded
 *    1 -> not a successful match
 *
 * Notes/Dependencies:
 *    if the input event string doesn't match the expected format, then
 *    tec_filter_event always returns 0.
 */

static int
tec_filter_walk(void *filter_elem, void *event_elem)
{
   int i;
   ev_slot *filter = (ev_slot*)filter_elem;
   ev_slot *event = (ev_slot*)event_elem;
   unsigned char *non_utf8_val_buf;
   int            non_utf8_val_len;
   int            match;

   /* There shouldn't be any empty filters in the list. */
   assert(filter->name != NULL);

   /* Event w/o a class?  Automatic mismatch. */
   if (event->name == NULL)
      return 1;

   /* 1st slots are both "Class"? */
   assert(strcmp(event->name, "Class") == 0 &&
          strcmp(event->name, filter->name) == 0);

   /* Make sure classes match. */
   if ( filter->value != NULL && strcmp(filter->value, event->value) != 0)
      return 1;

	/* IY26027 for AS/400 convert the event value to non utf8 because the 
	 * regular expressions in AS/400 are non utf8.
	 */
#ifdef OS400
	non_utf8_val_len = strlen(event->value) * TIS_FROM_UTF8_RATIO + 1;
	non_utf8_val_buf = (unsigned char *)malloc(non_utf8_val_len);
	NULL_CHECK(non_utf8_val_buf);
	tis_from_utf8(NULL, (unsigned char *)event->value, -1, non_utf8_val_buf,
					  non_utf8_val_len);
   match = eif_regexec(filter->regex, non_utf8_val_buf);
	free(non_utf8_val_buf);
#else
   match = eif_regexec(filter->regex, event->value);
#endif

   if ( (filter->regex != NULL) && !match )
      return 1;

	/*
	 * A match must be found for EVERY slot before we return 0 (success)
	 */
   ++filter;
   ++event;
   for (; filter->name != NULL; ++filter)
   {
      for (i = 0; event[i].name != NULL; ++i)
         if (strcmp(filter->name, event[i].name) == 0)
            break;
      /* No slot in the event w/ the name filter->name? */
      if (event[i].name == NULL)
         return 1;
      /* Mismatch on slot value? */
		if ( filter->value != NULL && strcmp(filter->value, event[i].value) != 0 )
			return 1;

		/* IY26027 for AS/400 convert the event value to non utf8 because the 
		 * regular expressions in AS/400 are non utf8.
		 */
#ifdef OS400
		non_utf8_val_len = strlen(event[i].value) * TIS_FROM_UTF8_RATIO + 1;
		non_utf8_val_buf = (unsigned char *)malloc(non_utf8_val_len);
		NULL_CHECK(non_utf8_val_buf);
		tis_from_utf8(NULL, (unsigned char *)event[i].value, -1, non_utf8_val_buf,
						  non_utf8_val_len);
		match = eif_regexec(filter->regex, non_utf8_val_buf);
		free(non_utf8_val_buf);
#else
	   match = eif_regexec(filter->regex, event[i].value);
#endif

		if ( (filter->value == NULL) && !match )
         return 1;
   }
   
   /* Everything matched. */
	return(0);
}

static void
tec_filter_delete(void *filter)
{
   deleteFilterOrEvent((ev_slot*)filter);
}

static void
tec_pre_filter_delete(void *filter_elem)
{
	NT_pre_filter	*pre_filter = (NT_pre_filter *) filter_elem;
	free_prefilter(pre_filter);
}

static void
tec_event_delete(void *ep)
{
	free(ep);
}

/*
 * Splits a comma separated list of entries. The number of entries is written
 * to num.  The split list is returned as an array of char* w/ *num entries.
 * It is the caller's responsibility to call free() on the returned pointer.
 */

char **
common_split_entries_base(char *s, int *num)
{
	char** buf;
	int i;

   *num = 0;
   if (s == NULL || *s == '\0')
   {
      buf = (char**)malloc(sizeof(char*));
      if (buf == NULL) fatal_memory_error();
      buf[0] = NULL;
		return buf;
   }

	/*
	 * Allocate a chunk of memory big enough for an array of char* of
    * size TECEIF_MAX_SERVER followed by the string all.
	 */
   buf = (char**)malloc(TECEIF_MAX_SERVERS*sizeof(char*) + strlen(s) + 1);
   if (buf == NULL) fatal_memory_error();
   strcpy(((char*)buf) + TECEIF_MAX_SERVERS*sizeof(char*), s);
   s = ((char*)buf) + TECEIF_MAX_SERVERS*sizeof(char*);

	for (i = 0; i < TECEIF_MAX_SERVERS; i++) {
		if ((buf[i] = strtok((i == 0) ? s : (char*) 0, ",")) == (char*) 0)
			break;
	}

	/*
	 * We ignore entries beyond TECEIF_MAX_SERVERS quietly
	 */
   
	*num = i;
	return buf;
}

char **
common_split_entries(tec_handle_common_t *ch, char* env, int* num)
{
   return common_split_entries_base(agent_getenv(ch->agent_config, env), num);
}

int
common_agent_match(tec_handle_common_t *ch, char *env_name, char *match_it)
{
	return(agent_match(ch->agent_config, env_name, match_it));
}

int
agent_match(AgentConfig *agent_info, char *env_name, char *match_it)
{
	char				*env_val;
	register char	*e, *m;

	if ((env_name == NULL) || (match_it == NULL))
		return(0);

	if ((env_val = agent_getenv(agent_info, env_name)) == NULL)
		return(0);

	for (e = env_val, m = match_it; (*m != '\0'); m++, e++) {
		if ((*e == '\0') || (toupper(*e) != *m))
			return(0);
	}

	return(1);
}


/*
 * Purpose:
 * free nt pre_filter
 * Description:
 */
static void
free_prefilter(NT_pre_filter *p)
{
	int	i;

	for (i = 0; p->event_src[i]; i++)
		free(p->event_src[i]);

#if defined(WIN32)
	for (i = 0; p->event_log[i]; i++)
		free(p->event_log[i]);

	for (i = 0; p->event_type[i]; i++)
		free(p->event_type[i]);
#endif

	free(p);
}

/*
 * Purpose:
 * determin if an event should be filtered out,
 * based upon the current prefilter.
 *
 * Description:
 * a match is that the event matches one of logs, one of ids and
 * one of types.
 *
 * Returns:
 * DISCARD_IT=1 -> discard the event
 * KEEP_IT=0 -> keep the event
 *
 */

int
common_pre_filter_event(tec_handle_common_t *ch, PreFilterEvent *event)
{
	AgentConfig				*agent_info;
	int						ret;

	agent_info = (ch && ch->agent_config) ? ch->agent_config : &default_agent_info;

	/*
	 * SL_Walk2 returns 0 if no match found, 1 otherwise
	 */
	ret = SL_Walk2(agent_info->tec_pre_filters, tec_pre_filter_walk, event);

	/*
	 * If filtering IN,  then no matching filter means to discard the event.
	 * If filtering OUT, then no matching filter means to send    the event.
	 */
	if (ret == 0)
		ret = (agent_info->pre_filter_in) ? FILTER_DISCARD : FILTER_KEEP;
	else
		ret = (agent_info->pre_filter_in) ? FILTER_KEEP : FILTER_DISCARD;

	return(ret);
}

/*
 * Purpose:
 * find a match with the pre_filter
 * Returns:
 * 1 -> no match
 * 0 -> match
 *
 */
static int
tec_pre_filter_walk(void *filter_elem, void *event_elem)
{
	NT_pre_filter	*p = (NT_pre_filter *) filter_elem;
	PreFilterEvent	*event = (PreFilterEvent *) event_elem;
   int				i;

	for (i = 0; (p->event_id[i] != -1); ++i) {
		if (event->id == p->event_id[i]) {
			break;
		}
	}

	if (i && (p->event_id[i] == -1)) {
		return 1;
	}

	if (event->source != NULL) {
		for (i = 0;  (p->event_src[i] != NULL); i++) {
			if (!strcmp(event->source, p->event_src[i]))
				break;
		}
	}

	if (i && (p->event_src[i] == NULL)) {
		return 1;
	}

#if defined(_WIN32)
	if (event->log != NULL) {
		for (i = 0;  (p->event_log[i] != NULL); i++) {
			if (!strcmp(event->log, p->event_log[i]))
				break;
		}
	}

   if (i && (p->event_log[i] == NULL)) {
      return 1;
   }

	if (event->type != NULL) {
		for (i = 0;  (p->event_type[i] != NULL); i++) {
			if (!strcmp(event->type, p->event_type[i]))
				break;
		}
	}

	if (i && (p->event_type[i] == NULL)) {
		return 1;
	}

#elif defined(NETWARE)
	for (i = 0; (p->event_severity[i] != -1); ++i) {
		if (event->severity == p->event_severity[i]) {
			break;
		}
	}

	if (i && p->event_severity[i] == -1) {
		return 1;
	}

	for (i = 0; (p->event_locus[i] != -1); ++i) {
		if (event->locus == p->event_locus[i]) {
			break;
		}
	}

	if (i && p->event_locus[i] == -1) {
		return 1;
	}

	for (i = 0; (p->event_class[i] != -1); ++i) {
		if (event->classname == p->event_class[i]) {
			break;
		}
	}

	if (i && p->event_class[i] == -1) {
		return 1;
	}
#endif

	return 0;
}

#if defined(WIN32) || defined(NETWARE)
static void
parse_pre_filter_ints(char *var, int valid_values[], int copy_array[],
					  int max_values)
{
	char	*str;
	int	c;
	int	v;
	int	val;
	int	copy_it;

	/*
	 * Parse and process each value token
	 */
	for (c = 0, str = strtok(NULL, ","); (str != NULL); str = strtok(NULL,",")) {
		val = atoi(str);

		/*
		 * If array of valid values provided, check value against the list.
		 * Otherwise, assume any value is valid.
		 */
		if (valid_values == NULL)
			copy_it = 1;
		else {
			copy_it = 0;

			for (v = 0; (valid_values[v] != INT_MAX); v++) {
				if (valid_values[v] == val) {
					copy_it = 1;
					break;
				}
			}
			if (!copy_it)
				fprintf(stderr,"PreFilter: s values '%d' not recognized.\n",val);
		}

		/*
		 * If the value is to be copied, do it if the max. # hasn't been reached.
		 */
		if (copy_it) {
			if (c < max_values) {
				copy_array[c] = val;
				c++;
			}
			else {
				fprintf(stderr, "PreFilter: number of %s values exceeds %d.\n",
						  var, max_values);
				fprintf(stderr, "           Value '%d' ignored.\n", val);
			}
		}
	}
	copy_array[c] = -1;
}
#endif

static void
parse_pre_filter(char *var, char *valid_values[], char *copy_array[],
					  int max_values)
{
	char	*str;
	int	c;
	int	v;
	int	copy_it;

	/*
	 * Parse and process each value token
	 */
	for (c = 0, str = strtok(NULL, ","); (str != NULL); str = strtok(NULL,",")) {
		/*
		 * If array of valid values provided, check value against the list.
		 * Otherwise, assume any value is valid.
		 */
		if (valid_values == NULL)
			copy_it = 1;
		else {
			copy_it = 0;

			for (v = 0; (valid_values[v] != NULL); v++) {
				if (!strcmp(str, valid_values[v])) {
					copy_it = 1;
					break;
				}
			}
			if (!copy_it)
				fprintf(stderr,"PreFilter: s values '%s' not recognized.\n",str);
		}

		/*
		 * If the value is to be copied, do it if the max. # hasn't been reached.
		 */
		if (copy_it) {
			if (c < max_values) {
				copy_array[c] = strdup(str);
				NULL_CHECK(copy_array[c]);
				c++;
			}
			else {
				fprintf(stderr, "PreFilter: number of %s values exceeds %d.\n",
						  var, max_values);
				fprintf(stderr, "           Value '%s' ignored.\n", str);
			}
		}
	}
	copy_array[c] = NULL;
}

/*
 * Purpose:
 * 	Make a NT prefilter
 *
 * Description:
 * 	Read in a line contains a NT prefilter info. and make a NT prefilter.
 *
 * Returns:
 * 	-1 -> syntax error.
 *  	0 -> OK.
 *
 */
static int
tec_add_prefilter(AgentConfig *agent_info, char *line)
{
#if defined(WIN32)
	char	*nt_logs[] = { "Application", "Security", "System", "Directory Service", "DNS Server", "File Replication Service", NULL };

	char	*nt_ev_types[] = { "Error", "Information", "Warning", "AuditSuccess", "AuditFailure", "Unknown", NULL };
#endif
	char				*s;
	char				*str;
	int				i;
	int				current_strlen;
	int				processed_strlen = 0;
	int				total_strlen;
	NT_pre_filter	*p;
	char				workline[512];

	if ((p = (NT_pre_filter *) calloc(1, sizeof(NT_pre_filter))) == NULL)
		fatal_memory_error();

   p->event_id[0] = -1;

#ifdef NETWARE
   p->event_severity[0] = -1;
   p->event_locus[0] = -1;
   p->event_class[0] = -1;
#endif

	total_strlen = strlen(line);
	strncpy(workline, line, total_strlen);
	workline[total_strlen] = '\0';

	while (processed_strlen < total_strlen) {
		s = strtok(workline, ";");
		current_strlen = strlen(s);
		str = strtok(s, "=");

		if (str != NULL) {
			if (!strcmp(str, SUBKEY_SOURCE)) { /* "Source" */
				parse_pre_filter(SUBKEY_SOURCE, NULL, p->event_src, MAX_EVENT_SRCS);
			}
#if defined(WIN32)
			else if ((!strcmp(str, SUBKEY_EVENT_ID))||(!strcmp(str, SUBKEY_GUI_EVENT_ID))) { /* "EventId" */
				parse_pre_filter_ints(SUBKEY_EVENT_ID, NULL, p->event_id, MAX_EVENT_IDS);
			}
			else if (!strcmp(str, SUBKEY_LOG)) { /* "Log" */
				parse_pre_filter(SUBKEY_LOG, nt_logs, p->event_log, MAX_EVENT_LOGS);
			}
			else if (!strcmp(str, SUBKEY_EVENT_TYPE)) { /* "EventType" */
				parse_pre_filter(SUBKEY_EVENT_TYPE, nt_ev_types, p->event_type, MAX_EVENT_TYPES);
			}
#elif defined(NETWARE)
			else if (!strcmp(str, SUBKEY_EVENT_ID)) { /* "EventId" */
				parse_pre_filter_ints(SUBKEY_EVENT_ID, NULL, p->event_id, MAX_EVENT_IDS);
			}
			else if (!strcmp(str, SUBKEY_SEVERITY)) { /* "Severity" */
				parse_pre_filter_ints(SUBKEY_SEVERITY, NULL, p->event_severity,
											MAX_EVENT_SEVERITIES);
			}
			else if (!strcmp(str, SUBKEY_LOCUS)) {
				parse_pre_filter_ints(SUBKEY_LOCUS, NULL, p->event_locus,
											MAX_EVENT_LOCUSES);
			}
			else if (!strcmp(str, SUBKEY_CLASS)) {
				parse_pre_filter_ints(SUBKEY_CLASS, NULL, p->event_class,
											MAX_EVENT_CLASSES);
			}
#endif
			else {
				for (i = 0; i <  current_strlen; i++) {
					if (!isspace(str[i])) {
#ifndef OS400
						fprintf(stderr, "PreFilter:contains illegal character.\n");
#else
						sendPgmMsg(TEC_MEM_FAIL, TEC_ADP_MSG_FILE_Q, " ", 0,
						           "*ESCAPE   ");
#endif

						free_prefilter(p);
						return -1;
					}
				}
			}
		}
		else
			fprintf(stderr, "PreFilter:contains illegal character.\n");

		processed_strlen += current_strlen + 1;

		if ((total_strlen - processed_strlen) > 0) {
			strncpy(workline,line + processed_strlen, total_strlen -processed_strlen);
			workline[total_strlen - processed_strlen] = '\0';
		}
	}
                /*
                 * Add filter to list of pre-filters.
                 */
#ifdef WIN32
        if (p->event_log[0] != NULL)
           {
#endif
                 if (SL_InsLast(agent_info->tec_pre_filters, (void *) p) == FALSE)
                    fatal_memory_error();
#ifdef WIN32
           }
#endif

	return 0;
}

#ifndef NO_UTF8
 /*
 * This is an init routine that is used by all flavors of adapters to
 * initialize some common data structure fields that are used during UTF-8
 * conversions and such.
 */
static void
tec_utf8_init(tec_handle_common_t *common_h)
{
   static int  firstTime = 1;
	char 			tablename[TIS_MAX_CS_NAME];
	char *env = NULL;

	/* Store adapter's local table; we'll need it later when doing tis_to_utf8 */

	tecadDEBUG_T(DEBUGFILE_AGENT_COMM,
					 "  storing adapter's local code set in tec_utf8_init ")

	/* Initialize tables	  */

	if (NULL != common_h) {
		/*
		 * Should this be set to the OS codeset or the adapter's default code set?
		 */
		common_h->adapter_local_cs_handle = NULL;
		tis_os_cs_name(tablename, TIS_MAX_CS_NAME);
		common_h->adapter_local_cs_handle = tis_cs_new(tablename);

      if (NULL == common_h->adapter_local_cs_handle)
		{
			/* throw an exception */
			tecadDEBUG_T(DEBUGFILE_AGENT_COMM,
							 "  ERROR: could not allocate code sets in utf8_init ")
		}

		common_h->tec_server_local_cs_handle = NULL;

		if (common_agent_match(common_h, VERSION_3_SERVER_COMPATIBILITY,"YES")){
			/* If they've specified a value for VERSION_SERVER_ENCODING in
			 * the conf file, we'll use it to initialize the
			 * tec_server_local_codeset variable in the data structure, which
			 * will be used later.
			 */
			env = agent_getenv(common_h->agent_config, VERSION_3_SERVER_ENCODING);

			if (NULL != env) {
				common_h->tec_server_local_cs_handle = tis_cs_new(env);
			}
		}
	}

	/*
	 * Tell FWK that we are a UTF-8 encoded process and,
	 * therefore, not to do conversion to UTF-8 (e.g. during
	 * CORBA marshalling for adapters for which marshalling occurs).
    */
   if (firstTime)
   {
      /*  Shouldn't the test of firstTime and installation of the
       *  UTF8 codeset be guarded by a mutex?  Yes, but TMF threads aren't
       *  preemptive so the test and setting of firstTime is atomic.
       */
      tis_cs_t utf8_cs_handle;
      firstTime = 0;
      utf8_cs_handle = tis_cs_new("UTF8");
      if (utf8_cs_handle == NULL)
		{
			/* throw an exception */
			tecadDEBUG_T(DEBUGFILE_AGENT_COMM,
							 "  ERROR: could not allocate code sets in utf8_init ")
		}
      else
      {
         tis_set_def_cs(utf8_cs_handle);
          tis_cs_free(utf8_cs_handle);
      }
   }
}

/* cleanup routine for utf8 data structures */
void
tec_utf8_cleanup(tec_handle_common_t *common_h)
{
	if (NULL != common_h) {
   	/* Free codeset handles */
		if (NULL != common_h->adapter_local_cs_handle)
			tis_cs_free(common_h->adapter_local_cs_handle);

		if (NULL != common_h->tec_server_local_cs_handle)
			tis_cs_free(common_h->tec_server_local_cs_handle);
	}
}

/* A utility routine to do conversion to utf8 - simply allocates
 * space for the converted message and calls the appropriate tis
 * API to do the conversion.
 */
char *
tec_to_utf8(char * message, tis_cs_t adapter_local_cs_handle)
{
	/* some temporary variables for UTF-8 conversions */
	int 				utf8_msg_len = 0;
	unsigned char	*utf8_msg_buf = NULL;

	/* Convert from adapter's local encoding to UTF-8.
	 * First, make sure that we're using the adapter's
	 * local table so that tis_to_utf8() will work correctly.
	 */

	tecadDEBUG_T(DEBUGFILE_AGENT_COMM, "  preparing to convert to UTF8 in tec_to_utf8")

	/* Convert the incoming event to UTF-8 from adapter's local encoding or
	 * convert from gateway's local encoding, if this is the tec_gateway
	 * (this code is used by both the tec_gateway and managed node adapters).
	 */
	if (NULL != message) {
		utf8_msg_len = strlen(message) * TIS_TO_UTF8_RATIO + 1;
		utf8_msg_buf = (unsigned char *)malloc(utf8_msg_len);
		NULL_CHECK(utf8_msg_buf);
		tis_to_utf8(adapter_local_cs_handle, (unsigned char *)message,
	  	         -1, utf8_msg_buf, utf8_msg_len);
	}
	else {
		/* This case is an error condition and should never happen;
		 * just in case it does, we'll send back a null string to avoid
		 * crashes that could occur if we returned a NULL ptr */
		utf8_msg_buf = (char *)malloc(sizeof(char));
		NULL_CHECK(utf8_msg_buf);
		utf8_msg_buf[0] = '\0';
	}

	return(utf8_msg_buf);
}

/* Puts a "magic" byte at the beginning of the event stream which
 * flags the event as UTF-8-encoded to the TEC server */
char *
tec_prepend_utf8_byte_to_msg(char *message)
{
   if (message[0] != '\x03' )
   {
	
   /* temporary variable for prepending UTF-8 magic byte */
	char				*msg_buf_with_utf8_byte = NULL;

	tecadDEBUG_T(DEBUGFILE_AGENT_COMM, " prepending magic utf8 byte ")

	if (NULL != message) {
		msg_buf_with_utf8_byte = (char*) malloc(strlen(message) + 2);
		NULL_CHECK(msg_buf_with_utf8_byte);
		msg_buf_with_utf8_byte[0] = TEC_IPC_UTF8_DELIMITOR;
		strcpy(msg_buf_with_utf8_byte + 1, message);
	}
	else {
		/* This case is an error condition and should never happen;
		 * just in case it does, we'll send back a null string to avoid
		 * crashes that could occur if we returned a NULL ptr */
		msg_buf_with_utf8_byte = (char *)malloc(sizeof(char));
		NULL_CHECK(msg_buf_with_utf8_byte);
		msg_buf_with_utf8_byte[0] = '\0';
	}
	return(msg_buf_with_utf8_byte);
   }
   else 
       return(message);
}

/* Converts the message to the TEC server's local encoding, as specified in
 * conf file (Pre37Server), for backwards compatibility with
 * old TEC servers. */
char *
tec_to_server_encoding(tec_handle_common_t *common_h, char *message)
{
	/* some temporary variables for UTF-8 conversions */
	int 				local_encoded_msg_len = 0;
	unsigned char	*local_encoded_msg_buf = NULL;

	if (NULL != message && NULL != common_h) {
		local_encoded_msg_len = strlen(message) * TIS_FROM_UTF8_RATIO + 1;
		local_encoded_msg_buf = (unsigned char *)malloc(local_encoded_msg_len);
		NULL_CHECK(local_encoded_msg_buf);

		if (NULL != common_h->tec_server_local_cs_handle) {
			/* If the adapter has a valid code set specified for
			 * Pre37ServerEncoding in the conf file, convert to it
			 * before sending.
			 */
			tecadDEBUG_T(DEBUGFILE_AGENT_COMM,
							 "  converting to tec server's local encoding")
			tis_from_utf8(common_h->tec_server_local_cs_handle,
							  (unsigned char*)message, -1, local_encoded_msg_buf,
							  local_encoded_msg_len);
		}
		else {
			/* Otherwise, use the adapter's local encoding as a fall
			 * back in the event that a bogus value (or nothing) is
			 * specified in the conf file.
	 		 */
			tecadDEBUG_T(DEBUGFILE_AGENT_COMM,
							 "  converting to adapter's local encoding")
			tis_from_utf8(common_h->adapter_local_cs_handle,
							  (unsigned char*)message, -1, local_encoded_msg_buf,
							  local_encoded_msg_len);
		}
	}
	else {
		/* This case is an error condition and should never happen;
		 * just in case it does, we'll send back a null string to avoid
		 * crashes that could occur if we returned a NULL ptr */
		local_encoded_msg_buf = (char *)malloc(sizeof(char));
		NULL_CHECK(local_encoded_msg_buf);
		local_encoded_msg_buf[0] = '\0';
	}
	return(local_encoded_msg_buf);
}

/* Utility routine to prep a message for transmission.
 * This determines whether or not to convert to the TEC server's local
 * encoding prior to transmission and whether or not to prepend
 * the "magic" byte to the event.
 */
char *
tec_prep_utf8_msg(tec_handle_common_t *common_h, char *message)
{
	if (common_agent_match(common_h, VERSION_3_SERVER_COMPATIBILITY, "YES")) {
		/* If we are sending to an OLD (version 3.6 or earlier)
  		 * TEC server, then we need to convert from UTF-8 to the
 		 * TEC server's local encoding before sending, since the old TEC
 		 * server isn't UTF-8-enabled. We cannot know, a priori, what the
 		 * TEC server's local encoding is; however, it can be specified
 		 * in the adapter's conf file (Pre37Server).
		 *
	 	 * Also, if sending to an OLD server, we won't prepend the
 		 * "magic" byte since old TEC servers cannot handle it.
 		 */
		message = tec_to_server_encoding(common_h, message);
	}
	else {
		/* We're sending to a new (version >= 4.0) Server, so we won't need to
		 * convert to the server's local encoding prior to transmission. Also,
		 * we will need to prepend the "magic" byte to the message, in order to
		 * flag the message as UTF-8 to the new server.
		 */
		message = tec_prepend_utf8_byte_to_msg(message);
	}
	return(message);
}

#ifdef NW4
char *cpl_getenv(char *name)
{
	return(getenv(name));
}
#endif 

#endif /* ifndef NO_UTF8 */

