/*
 * Component Name: EIF
 *
 * $Date: 2/18/02 15:31:06 $
 *
 * $Source: src/eif/evt_buffer.c, tec, tec_3.7.1 $
 *
 * $Revision: 1.24.1.23 $
 *
 * Description:
 *      TEC agent communication library for non-TME agents
 *
 * (C) COPYRIGHT TIVOLI Systems, Inc. 1994
 * Unpublished Work
 * All Rights Reserved
 * Licensed Material - Property of TIVOLI Systems, Inc.
 */

static int debugg=0;

#ifdef OS400
#include "QymaCpRight.h"
#include "qymadef.h"
#include <qcmdexc.cleinc>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <stdarg.h>
#include <assert.h>

#if defined(NETWARE) || defined(OS2)
#	include <share.h>
#endif

#if defined(WIN32) || defined(OS2)
#	include <direct.h>
#	include <io.h>
#else
#	include <limits.h>
#	include <unistd.h>
#endif

#if defined(NETWARE)
#	if defined(ANCIENT_SDK)
#		include <process.h>
#	else
#		include <nwthread.h>
#	endif
#	include <nwfile.h>
#endif

#if defined(WIN32)
#	define sleep(_s) Sleep((_s) * 1000)
#elif defined(OS2)
#	include <os2/os2def.h>
#	define INCL_DOSPROCESS yes
#	include <os2/bse.h>
#	define sleep(_s) DosSleep((_s) * 1000L)
#elif defined(NETWARE)
#	define sleep(_s) delay((_s) * 1000)
#endif

#ifdef WIN32
#	include <winsock.h>
#else
#	define HANDLE int
#endif

/* Are these REALLY needed?? */
#ifdef OS2
#	include <sys/socket.h>
#	include <netinet/in.h>
#	include <netdb.h>
#	include <signal.h>
#endif

#ifndef MAXPATHLEN
#  ifdef PATH_MAX
#     define MAXPATHLEN PATH_MAX
#  else
#     ifdef FILENAME_MAX
#	define MAXPATHLEN FILENAME_MAX
#     endif
#  endif
#endif

#include "tec_eif.h"
#include "slist.h"
#include "eif_common.h"
#include "eifdebug.h"

#ifndef NO_UTF8
#	include "tec_delims.h"
#endif

/*****************************************************************************
 **                        Platform Dependent Stuff                         **
 *****************************************************************************/

/*
 * Define a default buffer path
 */
#if defined(WIN32)
#	define DEFAULT_BUFFER_PATH "%SYSTEMROOT%\\SYSTEM32\\DRIVERS\\ETC\\TIVOLI\\TEC\\CACHE.DAT";
#elif defined(NETWARE)
#	define DEFAULT_BUFFER_PATH "SYS:ETC\\TIVOLI\\TEC\\CACHE.DAT"
#elif defined(OS400)
#	define DEFAULT_BUFFER_PATH "/etc/Tivoli/tec/"
#else
#	define DEFAULT_BUFFER_PATH "/etc/Tivoli/tec/cache"
#endif

/*
** With more recent NetWare SDKs, stat() and fstat() are macro aliases
** to stat_411() and fstat_411().  Since these symbols don't exist for
** NetWare 4.10, undefine these macros to force referencing the real
** stat() and fstat() symbols.
*/
#if defined(NETWARE)
#	if defined(stat)
#		undef stat
#	endif
#	if defined(fstat)
#		undef fstat
#	endif

/*
** With more recent NetWare SDKs, ctype functions are macro aliases that
** reference the symbol __ctype which is not present in NetWare 4.10.
** Undefine these macros and force referencing the actual ctype function
** symbols.
*/
#	if defined(isdigit)
#		undef isdigit
#	endif
#endif

/* sys_basename()
 *    This function treat path as the path name to a file, and returns
 *    the pointer into the string that points at the beginning of the
 *    base name of the file.  For example,
 *       sys_basename("dir1/dir2/basename")
 *    would return a pointer "basename" on UNIX systems.
 */
char* sys_basename(const char* path)
{
   char *p = (char*)(&path[strlen(path) - 1]);
#ifdef WIN32
   char *delimiters = ":/\\";
#elif defined(NETWARE)
   char *delimiters = "\\";
#else
   char *delimiters = "/";
#endif
   /* Did the user pass in an empty string? */
   assert (p >= path);

   for (; p > path; --p)
      if (strchr(delimiters, *p) != NULL)
         return p + 1;
   if (strchr(delimiters, *p) == NULL)
      return p;
   else
      return p + 1;
}

/* File creation flags used w/ sys_open_file().  These flags have the same
 * meaning as they do for UNIX open(), but we can't just use the UNIX flag
 * names directly because they have slightly different names on Win32.
 */
#ifdef WIN32
#define FIXUP_FOR_WIN32(x) _##x
#else
#define FIXUP_FOR_WIN32(x) x
#endif
#define SYS_O_RDONLY    FIXUP_FOR_WIN32(O_RDONLY)
#define SYS_O_WRONLY    FIXUP_FOR_WIN32(O_WRONLY
#define SYS_O_RDWR      FIXUP_FOR_WIN32(O_RDWR)
#define SYS_O_APPEND    FIXUP_FOR_WIN32(O_APPEND)
#define SYS_O_CREAT     FIXUP_FOR_WIN32(O_CREAT)
#define SYS_O_EXCL      FIXUP_FOR_WIN32(O_EXCL)
#define SYS_O_TRUNC     FIXUP_FOR_WIN32(O_TRUNC)

/* sys_open_file()
 *    Works just like the UNIX open() system call except that you don't
 *    get to specify the permissions with which the file is created if
 *    you specify the SYS_O_CREAT flag.  Use the SYS_O_* versions of the
 *    creation flags defined above because they work on all platforms.
 *
 *    Returns the open file handle on success and -1 on failure.
 */
HANDLE sys_open_file(const char *path, int flag)
{
	HANDLE					fh;

#ifdef WIN32
	SECURITY_ATTRIBUTES	sa;
   DWORD dwDesiredAccess = 0;
   DWORD dwCreationDistribution = 0;

	sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	sa.lpSecurityDescriptor = NULL;
	sa.bInheritHandle = TRUE;

   if      (flag & _O_RDONLY) { dwDesiredAccess=GENERIC_READ; }
   else if (flag & _O_WRONLY) { dwDesiredAccess=GENERIC_WRITE; }
   else if (flag & _O_RDWR)   { dwDesiredAccess=(GENERIC_READ|GENERIC_WRITE); }

   /* _O_APPEND not supported by Win32 natively. */
   if (flag & _O_APPEND) { return INVALID_HANDLE_VALUE; }

   if      ((flag & (_O_CREAT|_O_EXCL)) == (_O_CREAT|_O_EXCL))
      dwCreationDistribution = CREATE_NEW;
   else if ((flag & (_O_CREAT|_O_TRUNC)) == (_O_CREAT|_O_TRUNC))
      dwCreationDistribution = CREATE_ALWAYS;
   else if (flag & _O_CREAT)
      dwCreationDistribution = OPEN_ALWAYS;
   else if (flag & _O_TRUNC)
      dwCreationDistribution = TRUNCATE_EXISTING;
   else
      dwCreationDistribution = OPEN_EXISTING;
   
	fh = CreateFile(path,
                   dwDesiredAccess,
                   (FILE_SHARE_WRITE | FILE_SHARE_READ),
                   &sa,
                   dwCreationDistribution,
                   FILE_ATTRIBUTE_NORMAL,
                   NULL);

   if (fh == INVALID_HANDLE_VALUE)
      errno = (int) GetLastError();
#elif defined(OS2) || defined(NETWARE)
	/* open() chokes on OS/2 and NetWare if file is already open.  must
	 * use sopen() to open more than one at once.
	 */
   fh = sopen((path), O_BINARY | (flag), SH_DENYNO, S_IREAD|S_IWRITE);
#else
   /* UNIX */
   fh = open(path, flag, S_IRUSR|S_IWUSR);
#endif
	return(fh);
}

/* sys_close_file()
 *    Closes the file referenced by fd.  Returns TRUE/FALSE depending on
 *    success/failure.
 */
int sys_close_file(HANDLE fd)
{
#ifndef WIN32
   return close(fd) == 0;
#else
   return CloseHandle(fd);
#endif
}

/* sys_lock_file()
 *    Lock the entire file referenced by fd for exclusive access.  Returns
 *    TRUE/FALSE on success/failure.
 */
int sys_lock_file(HANDLE fd)
{
#ifndef WIN32
#if !defined(OS2) && !defined(NETWARE)
   /* UNIX */
   struct flock buflock;
   
	buflock.l_type = F_WRLCK;
	buflock.l_start = 0;
	buflock.l_whence = SEEK_SET;
	buflock.l_len = 0;

   return (fcntl(fd, F_SETLK, &buflock) >= 0);
#else
   /* BUG: OS2 & NETWARE - Is there really no locking? */
   return TRUE;
#endif
#else
   /* Win32 - Locking 2GB of length because there is no to EOF lock. */
	return LockFile(fd, 0, 0, 0x7FFFFFFF, 0);
#endif
}

/* sys_unlock_file()
 *    Unlocks the file referenced by fd.  Returns TRUE/FALSE on success/failure.
 */
int sys_unlock_file(HANDLE fd)
{
#ifndef WIN32
#if !defined(OS2) && !defined(NETWARE)
   /* UNIX */
   struct flock buflock;
   
	buflock.l_type = F_UNLCK;
	buflock.l_start = 0;
	buflock.l_whence = SEEK_SET;
	buflock.l_len = 0;

   return (fcntl(fd, F_SETLK, &buflock) >= 0);
#else
   /* BUG: OS2 & NETWARE - Is there really no locking? */
   return TRUE;
#endif
#else
   /* Win32 - Unlocking 2GB of length because there is no to EOF lock. */
	return UnlockFile(fd, 0, 0, 0x7FFFFFFF, 0);
#endif
}

/* sys_get_file_size2()
 *    Retrieves the size of the file in bytes and stores it in *size.
 *    Returns TRUE/FALSE on success/failure.
 */
int sys_get_file_size2(HANDLE fd, long *size)
{
   struct stat buffer_stat;
   eif_boolean retval;

   buffer_stat.st_size = -1;
#ifndef WIN32
   retval = (fstat(fd, &buffer_stat) == 0);
#else
   /* No stat on an fd in Win32. */
   retval = ((buffer_stat.st_size = GetFileSize(fd, NULL)) != 0xFFFFFFFF);
#endif
   *size = (long)buffer_stat.st_size;
   
   return retval;
}

/* sys_get_file_size()
 *    Retrieves the size of the file named by path in bytes and stores it
 *    in *size.  Returns TRUE/FALSE on success/failure.
 */
int sys_get_file_size(const char *path, long *size)
{
   HANDLE fh;
   int retval = FALSE;
   if ((fh = sys_open_file(path, O_RDONLY)) != INVALID_HANDLE_VALUE)
   {
      retval = sys_get_file_size2(fh, size);
      sys_close_file(fh);
   }
   return retval;
}

/* sys_lseek()
 *    Works just like lseek() in UNIX including the constants SEEK_SET,
 *    SEEK_CUR, and SEEK_END.  Returns -1 on error.
 */
long sys_lseek(HANDLE fd, long offset, int whence)
{
#ifndef WIN32
   return lseek(fd, offset, whence);
#else
   return (long)SetFilePointer(fd, offset, NULL, (whence == SEEK_SET ?
                                                  FILE_BEGIN :
                                                  (whence == SEEK_CUR ?
                                                   FILE_CURRENT :
                                                   FILE_END)));
#endif
}

/* sys_tell()
 *    Works just like tell() in UNIX.
 */
long sys_tell(HANDLE fd)
{
   return sys_lseek(fd, 0, SEEK_CUR);
}

/* sys_read()
 *    Attempts to read nbytes bytes from fd to buf.  Returns -1 on error,
 *    0 on EOF, and the number of bytes read otherwise.
 */
long sys_read(HANDLE fd, char *buf, long nbytes)
{
#ifndef WIN32
   return read(fd, buf, (size_t)nbytes);
#else
   DWORD bytesRead;
   BOOL  success = ReadFile(fd, (void*)buf, nbytes, &bytesRead, NULL);
   if (success) return (long)bytesRead;
   else         return -1;
#endif
}

/* sys_write()
 *    Attempt to write nbytes bytes from buf to fd.  Returns -1 on error
 *    and the number of bytes written otherwise.
 */
long sys_write(HANDLE fd, const char *buf, long nbytes)
{
#ifndef WIN32
   return write(fd, buf, (size_t)nbytes);
#else
   DWORD bytesWritten;
   BOOL  success = WriteFile(fd, (void*)buf, nbytes, &bytesWritten, NULL);
   if (success) return (long)bytesWritten;
   else         return -1;
#endif
}

/* sys_rename()
 *    Renames the file referenced by old_path to new_path.  Returns TRUE/FALSE
 *    on success/failure.
 */
int sys_rename(const char *old_path, const char *new_path)
{
   /* Using standard C library. */
   return rename(old_path, new_path) == 0;
}

/* sys_remove()
 *    Deletes the file referenced by the name "path".  Returns TRUE/FALSE
 *    on success/failure.
 */
int sys_remove(const char *path)
{
   /* Using standard C library. */
   return remove(path) == 0;
}

/* sys_truncate()
 *    Sets the size of the file referenced by fd to new_size bytes.  The
 *    function returns TRUE/FALSE on success/failure.  If the size of the
 *    file is less than new_size bytes, then it does nothing and returns
 *    FALSE.
 */
int sys_truncate(HANDLE fd, long newSize)
{
   long currSize;
   if (!sys_get_file_size2(fd, &currSize) || currSize < newSize)
      return FALSE;
   if (currSize == newSize)
      return TRUE;
#ifdef WIN32
   return sys_lseek(fd, newSize, SEEK_SET) == newSize && SetEndOfFile(fd);
#elif defined(OS2) || defined(NETWARE)
   return chsize(fd, newSize) == 0;
#else
   /* UNIX */
   return ftruncate(fd, newSize) == 0;
#endif
}

/*
 * sys_mkdir_r()
 *
 * Description:
 *      Recursively make the given directory.
 *      If the mkdir() works or the directory exists then return.
 *      If ENOENT is returned by mkdir(), something in the path
 *      doesn't exist, then remove the last node from the path and
 *      call sys_mkdir_r() to try and make that directory.
 *
 * Input:
 *      Dir - the directory path to create.
 */
static int sys_mkdir_r(const char *Dir) {
#if defined(NETWARE)
    /*
    ** Can't get meaningful errno from mkdir(), and can't open()
    ** and stat() a directory...
    **
    ** Unconditionally create the whole directory path, more
    ** elegant solutions notwithstanding...
    */

    char currPath[255];
    char* currElem = NULL;

    currPath[0] = '\0';
#ifdef NETWARE    
    currElem = strtok((char *)Dir, "\\");
#else    
    currElem = strtok(Dir, "\\");
#endif

    while (currElem != NULL) {
	if (currElem[strlen(currElem) - 1] == ':') {
	    strcat(currPath, currElem);
	    currElem = strtok(NULL, "\\");
	    continue;
	}

	strcat(currPath, currElem);
	strcat(currPath, "\\");
	mkdir(currPath);

	currElem = strtok(NULL, "\\");
    }

    return 0;

#else  /* !defined(NETWARE) */

	char    *DirDup;
	char    *PathEnd;
	int     RetVal = 0;
#ifdef OS400
	char	command[300];
#endif


#if !defined(WIN32) && !defined(OS2)
	if (mkdir(Dir, (S_IRWXU | S_IRGRP | S_IROTH)) != 0)
#else
	/*
	 * Don't try to make a directory that is actually a drive or partition name
	 */
	if (Dir[strlen(Dir)-1] == ':')
		return RetVal;

	if (mkdir(Dir) != 0)
#endif
	{
		switch (errno) {
			/*
			 * Part of path doesn't exist: create parent path then create Dir.
			 */
			case ENOENT:
				DirDup = strdup(Dir);
				NULL_CHECK(DirDup);

				if ((PathEnd = strrchr(DirDup, '/')) != NULL) {
					*PathEnd = 0; /* terminate shortened path */

					if (sys_mkdir_r(DirDup) == 0) {
#if !defined(WIN32) && !defined(OS2)
						if (mkdir(Dir, (S_IRWXU | S_IRGRP | S_IROTH)) != 0)
#else
						if (mkdir(Dir) != 0)
#endif
						{
							RetVal = -4;
						}
#ifdef OS400
						else {
							sprintf(command, "QSYS/CHGOWN OBJ('%s') NEWOWN(QSYS)",
							        Dir);
							QCMDEXC(command, strlen(command));
							chmod(Dir, S_IRWXU|S_IRWXG|S_IROTH|S_IXOTH);
						}
#endif
					}
					else {
						RetVal = -2;
					}
				}
				else {
					RetVal = -3; /* mkdir err creating first elem of Dir */
				}
				free(DirDup);
				break;

			case EEXIST:
#ifdef OS2
			case EACCES:
#endif
				RetVal = 0;
				break;

			default:
				RetVal = -1;  /* must be some sort of error */
				break;
		}
	}
#ifdef OS400
        else {
           sprintf(command, "QSYS/CHGOWN OBJ('%s') NEWOWN(QSYS)", Dir);
           QCMDEXC(command, strlen(command));
           chmod(Dir, S_IRWXU|S_IRWXG|S_IROTH|S_IXOTH);
        }
#endif

	return RetVal;
#endif /* NETWARE */
}


/*****************************************************************************
 **                  Event Cache File - Goals and Design                    **
 *****************************************************************************/
/*
 *   Goals:
 *   -----
 *   1. Don't send an event more than once, even if the adapter crashes.
 *   2. Don't lose events even if the adapter crashes.
 *   3. Consider the possibility of multiple adapters attempting to write/read
 *      from the same cache file simultaneously.
 *   4. Handle a cache file being larger than the limit specified in the
 *      adapter configuration file since the user might change the size
 *      between runs.
 *   5. Cached events are removed in a FIFO order.
 *   6. Old cache files should be coverted to the updated style whenever
 *      one is encountered.  Customers don't want to lose events when they
 *      upgrade.
 *
 *   What Ifs:
 *   --------
 *   1. Q. What happens when the size limit on the cache file is reached and
 *         we try to write a new event to the cache?
 *      A. Old events are removed until there is enough space for the new
 *         event.
 *
 *   2. Q. What happens when you try to cache an event that is too big to write
 *         to the cache file at all?
 *      A. The event is just discarded.
 *
 *   Cache File Format:
 *   -----------------
 *
 *   maxsz: XXXXXXXXXX\n
 *   head : XXXXXXXXXX\n
 *   tail : XXXXXXXXXX\n
 *   ........................event1 event2     \
 *   event3 event4 event5.................      |
 *   .....................................      |- Up to maxsz - 54 bytes.
 *   .....................................      |
 *   .....................................     /
 *
 *
 *   The format of the cache file is a circular array of up to size bytes,
 *   with head being the byte offset to the next event to send, and tail
 *   being the byte offset to the 1st byte of free space in the file.  The
 *   boundaries between events in the cache file are indicated by the
 *   terminating character TECAD_EVENT_END_CHAR at the end of each event.
 *
 *   The first three lines in the cache file all have a fixed size of 18
 *   bytes and record the following data:
 *      maxsz - The maximum size of the cache file.
 *      head  - The byte offset from the beginning of the file to the next
 *              event to send.  A value of 0 indicates an empty cache file.
 *      tail  - The byte offset from the beginning of the file to the first
 *              byte of free space in the file.
 *
 *****************************************************************************/

/*
cache_constructor
cache_destructor
cache_convert_file
cache_resize_file
cache_open_file
cache_is_open
cache_file_name
cache_max_size
cache_is_empty
cache_free_space
cache_get_curr_event
cache_step_file
cache_curr_step_no
cache_done_with_ev_cached_at_open
cache_append_event
cache_sync
cache_truncate
cache_dump
        

void cache_constructor(tec_event_cache_t *pEC)
void cache_destructor(tec_event_cache_t *pEC)
static int cache_convert_file(tec_event_cache_t *pEC)
int cache_resize_file(tec_event_cache_t *pEC, long new_max_size)
eif_boolean cache_open_file(tec_event_cache_t *pEC,
                            const char *file_path,
                            long max_size)
eif_boolean cache_is_open(const tec_event_cache_t *pEC)
const char* cache_file_name(const tec_event_cache_t *pEC)
long cache_max_size(const tec_event_cache_t *pEC)
eif_boolean cache_is_empty(const tec_event_cache_t *pEC)
long cache_free_space(const tec_event_cache_t *pEC)
eif_boolean cache_get_curr_event(tec_event_cache_t *pEC, char **event)
eif_boolean cache_step_file(tec_event_cache_t *pEC)
long cache_curr_step_no(const tec_event_cache_t *pEC)
eif_boolean cache_done_with_ev_cached_at_open(const tec_event_cache_t *pEC)
eif_boolean cache_append_event(tec_event_cache_t *pEC, const char *event)
eif_boolean cache_sync(tec_event_cache_t *pEC)
eif_boolean cache_truncate(tec_event_cache_t *pEC)
eif_boolean cache_dump(const tec_event_cache_t *pEC, 
                       cache_dump_func_t callback,
                       void *callback_data,
                       cache_dump_mode_t dmode)
*/

/*****************************************************************************
 **                  Functions for Managing a Cache File                    **
 *****************************************************************************/
/*
 * These functions all take a pointer to a tec_event_cache_t structure.  The
 * intent is that all cache file manipulation is accomplished with these
 * functions, thereby giving some level of data abstraction.
 */

/* Macro supporting the modular arithmetic necessary for doing a circular
 * cache.  Note that because of the cache header, we can't simply use the
 * mod operator '%'. */
#ifndef INC_FILE_PTR
#define INC_FILE_PTR(fp, offset, maxsz) \
{ \
   (fp) += (offset); \
   if ((fp) >= (maxsz)) \
      (fp) = CACHE_HDR_LEN + ((fp) - (maxsz)); \
}
#else
#error INC_FILE_PTR already defined.
#endif

/* scanf/printf format for cache file header. */
const char CACHE_HDR_FORMAT[] = 
   "maxsz: %10ld\n"
   "head : %10ld\n"
   "tail : %10ld\n";

/* Length in bytes of cache header. */
#define CACHE_HDR_LEN 54

/* Minimum on the maximum size setting of a cache file. */
const long MIN_CACHE_SZ = 8 * 1024;

/* Size of the chunk of data we read/write from/to the cache file. */
const long CACHE_RDWR_SZ = 32 * 1024;

#ifdef DEBUG
/* tecadDEBUG_support_func()
 *    Debug printing support function declared in common/eifdebug.h.
 */
const char *tecadDEBUG_dbgFile = NULL;
const char *tecadDEBUG_srcFile = NULL;
int         tecadDEBUG_srcLine = 0;
void tecadDEBUG_support_func(const char *format, ...)
{
   FILEOPEN(tecadDEBUG_dbgFile);
   if (fp) 
   {
      va_list ap;
      fprintf(fp, "DEBUG in %s:%i -- ", tecadDEBUG_srcFile, tecadDEBUG_srcLine);
      va_start(ap, format);
      vfprintf(fp, format, ap);
   }
   FILECLOSE(fp);
}
#endif /* DEBUG */

/* cache_constructor()
 *    Initializes a tec_event_cache_t "object."
 */
void cache_constructor(tec_event_cache_t *pEC)
{
   pEC->file_path = NULL;
   pEC->fd = INVALID_HANDLE_VALUE;
   pEC->buf = NULL;
   pEC->wbuf = NULL;
}

/* cache_destructor()
 *    Closes the cache and frees associated resources.  It returns the
 *    tec_event_cache_t structure to the state it has right after a
 *    call to cache_constructor().
 */
void cache_destructor(tec_event_cache_t *pEC)
{
   if (pEC->file_path != NULL) free(pEC->file_path);
   if (pEC->fd != INVALID_HANDLE_VALUE)
   { sys_unlock_file(pEC->fd); sys_close_file(pEC->fd); }
   if (pEC->buf != NULL) free(pEC->buf);
   if (pEC->wbuf != NULL) free(pEC->wbuf);
   cache_constructor(pEC);
}

/* cache_convert_file()
 *    Convert from old style to new style cache file.  Returns 0 on success,
 *    1 when the conversion fails but the original file is still intact,
 *    and 2 when conversion fails and the state of the original file is
 *    undefined.  This function assumes the context in which it's called in
 *    cache_open_file() and is really a "private" support function for
 *    cache_open_file.
 */
static int cache_convert_file(tec_event_cache_t *pEC)
{
#  define RD_BUF_SZ  8 * 1024
   char buf[RD_BUF_SZ];
   long bytesLeft;
   long newMaxsz, newHead, newTail;
   long rpos, bytesToRead;
   int retval = 1;

   tecadDEBUG_pf(("cache_convert_file(): pEC->file_path = (%s)\n",
                  pEC->file_path));

   if (!sys_get_file_size2(pEC->fd, &bytesLeft))
      goto EXIT;
   tecadDEBUG_pf(("cache_convert_file(): bytesLeft = %lu\n", bytesLeft));
   newMaxsz = bytesLeft + CACHE_HDR_LEN + 8*1024;
   newHead = CACHE_HDR_LEN;
   newTail = bytesLeft + CACHE_HDR_LEN;

   /* Shift everything down CACHE_HDR_LEN bytes in the file. */
   while (bytesLeft > 0)
   {
      bytesToRead = (bytesLeft >= RD_BUF_SZ ? RD_BUF_SZ : bytesLeft);
      rpos = (bytesLeft >= RD_BUF_SZ ? bytesLeft - RD_BUF_SZ : 0);
      if (sys_lseek(pEC->fd, rpos, SEEK_SET) != rpos ||
          sys_read(pEC->fd, buf, bytesToRead) != bytesToRead)
      {
         tecadDEBUG_pf(("cache_convert_file(): seek and read failed, bytesLeft = %lu, rpos = %ld\n", bytesLeft, rpos));
         goto EXIT;
      }
      /* About to modify cache file, so it is indeterminate state. */
      retval = 2;
      if (sys_lseek
          (pEC->fd, rpos + CACHE_HDR_LEN, SEEK_SET) != rpos + CACHE_HDR_LEN ||
          sys_write(pEC->fd, buf, bytesToRead) != bytesToRead)
      {
         tecadDEBUG_pf(("cache_convert_file(): seek and write failed, bytesLeft = %lu, rpos = %ld\n", bytesLeft, rpos));
         goto EXIT;
      }
      bytesLeft -= bytesToRead;
   }

   /* Write the header to the file. */
   {
      char hdr[CACHE_HDR_LEN + 1];
      sprintf(hdr, CACHE_HDR_FORMAT, newMaxsz, newHead, newTail);
      if (sys_lseek(pEC->fd, 0, SEEK_SET) != 0 ||
          sys_write(pEC->fd, hdr, CACHE_HDR_LEN) != CACHE_HDR_LEN)
      {
         tecadDEBUG_pf(
            ("cache_convert_file(): seek and write of header failed\n"));
         goto EXIT;
      }
   }
   
   pEC->maxsz = newMaxsz;
   pEC->head = newHead;
   pEC->tail = newTail;
   retval = 0;

  EXIT:
   return retval;
}

/* cache_resize_file()
 *    Resize the cache file from it's current max size to new_max_size. If
 *    new_max_size is less than 8K, then 8K is used for new_max_size instead.
 *    The function works by rebuilding a cache file of max size new_max_size by
 *    calling cache_append_event() repeatedly.  This function returns
 *    0 on success, 1 when the resize fails but the original file is still
 *    intact, and 2 when conversion fails and the state of the original file
 *    is undefined. This function assumes that the cache file has been
 *    opened by cache_open_file().
 *
 *    ** WARNING **
 *    Resizing the cache file erases all record of old consumed data that
 *    could be stepped through with cache_dump().  Only the most recent
 *    data that will fit in a file of new_max_size is preserved.
 */
int cache_resize_file(tec_event_cache_t *pEC, long new_max_size)
{
   int retval = 1;
   char *event = NULL;
   char *tmp_path = NULL;
   tec_event_cache_t tEC;
   long orig_tail_on_open = pEC->tail_on_open;
   long new_tail_on_open = 0;

   /* Catch the user passing in something ridiculous for the max size. */
   if (new_max_size < MIN_CACHE_SZ)
      new_max_size = MIN_CACHE_SZ;

   /* Catch special case where user is not actually changing the size. */
   if (new_max_size == pEC->maxsz) return 0;

   cache_constructor(&tEC);

   /* Try to open the new cache file. */
   {
      tmp_path = (char*)malloc(strlen(pEC->file_path) + 8);
      NULL_CHECK(tmp_path);
      sprintf(tmp_path, "%s.TMPXYZ", pEC->file_path);
      tecadDEBUG_pf(("cache_resize_file(): new_max_size = %lu, tmp_path = %s\n",
                     new_max_size, tmp_path));
      if (!cache_open_file(&tEC, tmp_path, new_max_size) ||
          !cache_is_empty(&tEC) ||
          cache_max_size(&tEC) != new_max_size)
      {
         tecadDEBUG_pf(
            ("cache_resize_file(): opening empty cache file failed.\n"));
         goto EXIT;
      }
   }

   /* Step through old cache file writing events to the new cache file. */
   {
      eif_boolean getCurrRes;
      long last_tail_on_open = pEC->tail_on_open;
      while ((getCurrRes = cache_get_curr_event(pEC, &event)) && event != NULL)
      {
         if (!cache_append_event(&tEC, event) ||
             !cache_step_file(pEC))
         {
            pEC->tail_on_open = orig_tail_on_open;
            goto EXIT;
         }
         if (last_tail_on_open != pEC->tail_on_open)
         {
            new_tail_on_open = tEC.tail;
            last_tail_on_open = pEC->tail_on_open;
         }
      }
      if (!getCurrRes || !cache_sync(&tEC))
         goto EXIT;
   }
   
   /* Rename stuff so that the new cache file takes the place of the
    * old one, and remove the old cache file. */
   /* BUG: unlocking, removing, and renaming needs to be an atomic
    *      operation.  Just living w/ the bug because getting around
    *      it is too hard. This is why there are 3 return values for
    *      this function. */
   retval = 2;
   sys_unlock_file(pEC->fd);
   sys_close_file(pEC->fd);
   pEC->fd = INVALID_HANDLE_VALUE;
   if (!sys_remove(pEC->file_path) || !sys_rename(tmp_path, pEC->file_path))
      goto EXIT;

   /* Copy stuff from tEC into *pEC. */
   pEC->fd = tEC.fd;
   tEC.fd = INVALID_HANDLE_VALUE;
   pEC->maxsz = tEC.maxsz;
   pEC->head = tEC.head;
   pEC->tail = tEC.tail;
   pEC->tail_on_open = new_tail_on_open;

   /* Forget about any cached data since we just changed files. */
   pEC->head_at_read = pEC->head;
   pEC->bytes_read = 0;
   pEC->curr_pos = pEC->buf;
   if (pEC->curr_pos != NULL)
      pEC->curr_pos[0] = '\0';

   retval = 0;

  EXIT:
   cache_destructor(&tEC);
   sys_remove(tmp_path);
   if (event != NULL) free(event);
   if (tmp_path != NULL) free(tmp_path);
   return retval;
}

/* cache_open_file()
 *    Opens the cache file named by file_path, and reads the values of the
 *    state variables maxsz, head, and tail from the beginning of the file.
 *    If the file doesn't have a header containing these state variables,
 *    then the presumption is that it's an old-style cache file, and it is
 *    converted by calling cache_convert_file().  A lock is established on
 *    the entire file.  If the cache file doesn't exist, then it is created
 *    along with any non-existent directories in file_path.
 *
 *    If the cache file is newly created cache file, then max_size is used as
 *    the maximum size for the new cache file.  If max_size is specified as
 *    something less than 8K, then 8K is used for max_size instead.
 *
 *    If the cache file is created by converting it from an old style cache
 *    file, then the maximum size set for the cache file is the size of
 *    the cache file after conversion plus 8K.
 *
 *    When opening a "normal" cache file, max_size is ignored in preference
 *    to the "maxsz" field stored in the header of the cache file.
 *
 *    It is an error to call this function if the cache file is already open.
 *
 *    The function returns TRUE/FALSE on success/failure.
 */
eif_boolean cache_open_file(tec_event_cache_t *pEC,
                            const char *file_path,
                            long max_size)
{
   eif_boolean retval = FALSE;
   long   nbytes;
   char   hdr[CACHE_HDR_LEN + 1];
   char  *dir_path = NULL;
   char  *basename = NULL;

   NULL_CHECK(file_path);
   /* Is the cache file already open? */
   assert(pEC->file_path == NULL);

   pEC->fd = INVALID_HANDLE_VALUE;

   /* Catch the user passing in something ridiculous for the max size. */
   if (max_size < MIN_CACHE_SZ)
      max_size = MIN_CACHE_SZ;

   /* Catch special case of an empty file_path. */
   if (file_path[0] == '\0') goto EXIT;

   tecadDEBUG_pf(("cache_open_file(): Attempt to open %s\n", file_path));

   if ((pEC->fd = sys_open_file(file_path, SYS_O_CREAT|SYS_O_RDWR)) ==
       INVALID_HANDLE_VALUE)
   {
      /* Try creating directories in path to cache file to see if that
       * is why we couldn't open it. */
      dir_path = strdup(file_path);
      basename = sys_basename(dir_path);
      tecadDEBUG_pf(
         ("cache_open_file(): Making dir where cache file lives.\n"));
      if (basename == dir_path) goto EXIT;
      *(basename - 1) = '\0';
      if (sys_mkdir_r(dir_path) != 0)
      {
         tecadDEBUG_pf(
            ("cache_open_file(): Making dir for cache file failed.\n"));
         goto EXIT;
      }


      if ((pEC->fd = sys_open_file(file_path, SYS_O_CREAT|SYS_O_RDWR)) ==
          INVALID_HANDLE_VALUE)
      {
         tecadDEBUG_pf(
            ("cache_open_file(): Couldn't open cache file at all.\n"));
         free(dir_path);
         return FALSE;
      }
   }

   tecadDEBUG_pf(("cache_open_file(): 1.\n"));

   if (!sys_lock_file(pEC->fd))
   {
      tecadDEBUG_pf(("cache_open_file(): Locking cache file failed.\n"));
      goto EXIT;
   }

   tecadDEBUG_pf(("cache_open_file(): 2.\n"));

   pEC->file_path = strdup(file_path);
   NULL_CHECK(pEC->file_path);
   
   tecadDEBUG_pf(("cache_open_file(): 3.\n"));

   nbytes = sys_read(pEC->fd, hdr, CACHE_HDR_LEN);
   hdr[CACHE_HDR_LEN] = '\0';
   if (nbytes == -1) goto EXIT;
   tecadDEBUG_pf(("cache_open_file(): 4.\n"));
   if (nbytes == 0)
   {
      /* New cache file. */
      tecadDEBUG_pf(("cache_open_file(): new cache file\n"));
      pEC->maxsz = max_size;
      pEC->head = 0;
      pEC->tail = 0;

      sprintf(hdr, CACHE_HDR_FORMAT, pEC->maxsz, pEC->head, pEC->tail);
      if (sys_lseek(pEC->fd, 0, SEEK_SET) != 0 ||
          sys_write(pEC->fd, hdr, CACHE_HDR_LEN) != CACHE_HDR_LEN)
      {
         tecadDEBUG_pf(("cache_open_file(): Writing header failed.\n"));
         goto EXIT;
      }
   }
   else
   if (nbytes != CACHE_HDR_LEN ||
       sscanf(hdr, CACHE_HDR_FORMAT, &pEC->maxsz, &pEC->head, &pEC->tail) != 3)
   {
      /* If the cache file doesn't have a legal header, assume it's an old cache
       * file and try to convert it. */
      tecadDEBUG_pf(("cache_open_file(): calling cache_convert_file()\n"));
      if (cache_convert_file(pEC) != 0)
      {
         tecadDEBUG_pf(("cache_open_file(): cache_convert_file() failed.\n"));
         goto EXIT;
      }
   }
   tecadDEBUG_pf(("cache_open_file(): 5.\n"));
   pEC->tail_on_open = (pEC->head == 0 ? 0 : pEC->tail);
   pEC->head_at_read = 0;
   pEC->bytes_read = 0;
   pEC->buf = NULL;
   pEC->curr_pos = NULL;
   pEC->curr_step_no = 0;
   pEC->tail_to_write = 0;
   pEC->bytes_to_write = 0;
   pEC->wbuf = NULL;
   pEC->read_only = FALSE;
   retval = TRUE;

  EXIT:
   if (!retval)
   {
      if (pEC->file_path != NULL) free(pEC->file_path);
      pEC->file_path = NULL;
      if (pEC->fd != INVALID_HANDLE_VALUE)
      { sys_unlock_file(pEC->fd); sys_close_file(pEC->fd); }
      pEC->fd = INVALID_HANDLE_VALUE;
   }
   free(dir_path);
   return retval;
}

/* cache_is_open()
 *    Returns TRUE if the cache file is currently open, and FALSE otherwise.
 */
eif_boolean cache_is_open(const tec_event_cache_t *pEC)
{
   return pEC->file_path != NULL;
}

/* cache_file_name()
 *    Returns the file name passed to cache_open_file() when the cache file
 *    was opened.
 */
const char* cache_file_name(const tec_event_cache_t *pEC)
{
   assert(cache_is_open(pEC));
   return pEC->file_path;
}

/* cache_max_size()
 *    Returns the maximum size of the cache file in bytes.  Assumes the cache
 *    file has been opened by cache_open_file().
 */
long cache_max_size(const tec_event_cache_t *pEC)
{
   assert(cache_is_open(pEC));
   return pEC->maxsz;
}

/* cache_is_empty()
 *    Returns TRUE/FALSE depending on whether or not the cache file is empty.
 */
eif_boolean cache_is_empty(const tec_event_cache_t *pEC)
{
   assert(cache_is_open(pEC));
   return (pEC->head == 0);
}

/* cache_free_space()
 *    Returns the number of bytes of free space left in the cache for
 *    caching events.
 */
long cache_free_space(const tec_event_cache_t *pEC)
{
   assert(cache_is_open(pEC));
   if      (pEC->head == 0)
      return pEC->maxsz - CACHE_HDR_LEN;
   else if (pEC->head < pEC->tail)
      return pEC->maxsz - (pEC->tail - pEC->head) - CACHE_HDR_LEN;
   else
      return pEC->head - pEC->tail;
}

/* cache_get_curr_event()
 *    Reads the current event from the event cache file.  Returns TRUE if the
 *    read was successful and FALSE otherwise.  If TRUE is returned and
 *    there was a current event (i.e., not end of cache) then *event points
 *    to a dynamically allocated buffer containing the null-terminated event
 *    data.  This data _includes_ the terminating char TECAD_EVENT_END_CHAR.
 *    It is the caller's responsibility to delete this buffer using free().  If
 *    the end of the cache has been reached, then *event is set to NULL.  If
 *    FALSE is returned then *event is set to NULL.
 *
 *    This function does not advance through the cache file, immediately making
 *    another call to cache_get_curr_event() will return the same event.
 */    
eif_boolean cache_get_curr_event(tec_event_cache_t *pEC, char **event)
{
   char *eventEnd;

   assert(cache_is_open(pEC));

   *event = NULL;
   if (cache_is_empty(pEC)) return TRUE;

   if (pEC->buf == NULL)
   {
      tecadDEBUG_pf(("cache_get_curr_event(): allocating buffer\n"));
      pEC->head_at_read = pEC->head;
      pEC->bytes_read = 0;
      pEC->buf = (char*)malloc(CACHE_RDWR_SZ + 1);
      NULL_CHECK(pEC->buf);
      pEC->curr_pos = pEC->buf;
      pEC->curr_pos[0] = '\0';
   }

   if ((eventEnd = strchr(pEC->curr_pos, TECAD_EVENT_END_CHAR)) == NULL)
   {
      /* The read buffer doesn't contain the end of the next event so
       * we've got to read in another chunk of data. */
      long bytesToRead;
      long readOffset;
      long bytesLeft = pEC->bytes_read - (pEC->curr_pos - pEC->buf);
      tecadDEBUG_pf(("cache_get_curr_event(): in code to read more data\n"));
      tecadDEBUG_pf(("cache_get_curr_event(): bytesLeft = %ld\n", bytesLeft));
      if (bytesLeft == CACHE_RDWR_SZ)
         /* No TECAD_EVENT_END_CHAR in a full block of data. Bad cache file. */
         goto ERROR_EXIT;

      /* Shift remaining data up in the buffer, and prepare for reading
       * more data. */
      memmove(pEC->buf, pEC->curr_pos, bytesLeft);
      INC_FILE_PTR(pEC->head_at_read, pEC->bytes_read - bytesLeft, pEC->maxsz);
      pEC->bytes_read = bytesLeft;
      pEC->curr_pos = pEC->buf;
      pEC->curr_pos[bytesLeft] = '\0';

      /* At this point the beginning of our buffer should coincide with the
       * head file pointer. */
      assert(pEC->head_at_read == pEC->head);

      /* Sync write buffer w/ disk if it's non-empty since the write buffer
       * may contain data we're about to try to read. */
      if (pEC->bytes_to_write > 0 && !cache_sync(pEC))
      {
         tecadDEBUG_pf(("cache_get_curr_event(): synching w/ disk failed.\n"));
         return FALSE;
      }

      readOffset = pEC->head;
      INC_FILE_PTR(readOffset, pEC->bytes_read, pEC->maxsz);
      bytesToRead = CACHE_RDWR_SZ - pEC->bytes_read;
      tecadDEBUG_pf(("cache_get_curr_event(): readOffset = %ld, bytesToRead = %ld, head = %lu, tail = %lu\n", readOffset, bytesToRead, pEC->head, pEC->tail));
      if (readOffset == pEC->tail && pEC->head != pEC->tail)
         /* Exhausted data in cache w/o finding a TECAD_EVENT_END_CHAR to
          * terminate the last event.  Bad cache file. */
         goto ERROR_EXIT;
      if (readOffset >= pEC->tail)
      {
         long bytesToEnd = pEC->maxsz - readOffset;
         if (bytesToRead < bytesToEnd)
            bytesToEnd = bytesToRead;
         tecadDEBUG_pf(
            ("cache_get_curr_event(): bytesToEnd = %ld\n", bytesToEnd));
         if (sys_lseek(pEC->fd, readOffset, SEEK_SET) != readOffset ||
             sys_read
             (pEC->fd, &pEC->buf[pEC->bytes_read], bytesToEnd) != bytesToEnd)
         {
            tecadDEBUG_pf(
               ("cache_get_curr_event(): 1. seek and read failed.\n"));
            goto ERROR_EXIT;
         }
         bytesToRead -= bytesToEnd;
         pEC->bytes_read += bytesToEnd;
         readOffset = CACHE_HDR_LEN;
      }
      tecadDEBUG_pf(("cache_get_curr_event(): readOffset = %ld, bytesToRead = %ld, head = %lu, tail = %lu\n", readOffset, bytesToRead, pEC->head, pEC->tail));
      if (bytesToRead > 0 && readOffset < pEC->tail)
      {
         if (bytesToRead > pEC->tail - readOffset)
            bytesToRead = pEC->tail - readOffset;
         if (sys_lseek(pEC->fd, readOffset, SEEK_SET) != readOffset ||
             sys_read
             (pEC->fd, &pEC->buf[pEC->bytes_read], bytesToRead) != bytesToRead)
         {
            tecadDEBUG_pf(
               ("cache_get_curr_event(): 2. seek and read failed.\n"));
            goto ERROR_EXIT;
         }
         pEC->bytes_read += bytesToRead;
      }
      pEC->buf[pEC->bytes_read] = '\0';

      tecadDEBUG_pf(("cache_get_curr_event(): pEC->bytes_read = %ld\n",
                     pEC->bytes_read));

      if ((eventEnd = strchr(pEC->curr_pos, TECAD_EVENT_END_CHAR)) == NULL)
         goto ERROR_EXIT;
   }

   *event = (char*)malloc(eventEnd - pEC->curr_pos + 2);
   NULL_CHECK(*event);
   memcpy(*event, pEC->curr_pos, eventEnd - pEC->curr_pos + 1);
   (*event)[eventEnd - pEC->curr_pos + 1] = '\0';
   return TRUE;

  ERROR_EXIT:
   /* If something went wrong, forget about any cached data. */
   *event = NULL;
   pEC->head_at_read = pEC->head;
   pEC->bytes_read = 0;
   pEC->curr_pos = pEC->buf;
   pEC->curr_pos[0] = '\0';
   return FALSE;
}

/* cache_step_file()
 *    This function steps to the next event in the cache file.  The function
 *    does not update the file on disk and you must call cache_sync() to
 *    sync the disk file with the step.  For perfomance reasons you might
 *    not want to sync w/ the disk on every call to cache_step_file().  The
 *    function returns TRUE/FALSE on success/failure.  If FALSE is returned,
 *    the set of events stored on disk hasn't been corrupted.  It is an error
 *    to call this function if the cache is empty.
 */
eif_boolean cache_step_file(tec_event_cache_t *pEC)
{
   char *eventEnd;
   long eventLen;
   long newHead;
   
   assert(cache_is_open(pEC));
   assert(!cache_is_empty(pEC));
   
   if (pEC->buf == NULL ||
       (eventEnd = strchr(pEC->curr_pos, TECAD_EVENT_END_CHAR)) == NULL)
   {
      char *event;
      if (!cache_get_curr_event(pEC, &event))
         return FALSE;
      free(event);
      eventEnd = strchr(pEC->curr_pos, TECAD_EVENT_END_CHAR);
      assert(eventEnd != NULL);
   }

   eventLen = eventEnd - pEC->curr_pos + 1;
   newHead = pEC->head;
   INC_FILE_PTR(newHead, eventLen, pEC->maxsz);
   if (newHead == pEC->tail)
      newHead = 0;

   /* Update cache state in memory only. */
   pEC->head = newHead;
   if (pEC->head == 0 || pEC->head == pEC->tail_on_open)
      /* Note that we just consumed all events present at the time we
       * opened the cache file. */
      pEC->tail_on_open = 0;
   if (pEC->head > 0)
      pEC->curr_pos += eventLen;
   else
   {
      /* Exhausted cache. Set up read buffer for the next call of
       * cache_get_curr_event() after more events have been added to the
       * cache file. */
      pEC->head_at_read = pEC->tail;
      pEC->bytes_read = 0;
      pEC->curr_pos = pEC->buf;
      pEC->curr_pos[0] = '\0';
   }
   ++pEC->curr_step_no;
   
   return TRUE;
}

/* cache_curr_step_no()
 *    Returns the number of successful calls to cache_step_file() that have
 *    been made since the cache file was opened w/ cache_open_file().  You
 *    might need this info since cache_append_event() can make calls
 *    to cache_step_file() that you don't ask for directly.
 */
long cache_curr_step_no(const tec_event_cache_t *pEC)
{
   assert(cache_is_open(pEC));
   return pEC->curr_step_no;
}

/* cache_done_with_ev_cached_at_open()
 *    This function returns true if cache_step_file() has been called enough
 *    times to step over all the events present in the cache file at the
 *    time it was opened w/ cache_open_file().
 */
eif_boolean cache_done_with_ev_cached_at_open(const tec_event_cache_t *pEC)
{
   assert(cache_is_open(pEC));
   return pEC->tail_on_open == 0;
}

/* cache_append_event()
 *    This function appends the event data of length strlen(event) pointed to
 *    by event to the end of the list in the cache file.  If the event
 *    is too big to fit in the cache file at all (for example, trying to
 *    write a 12k event to a 8k cache file) then nothing is done and
 *    TRUE is returned.  If the cache file is big enough to hold the event,
 *    but there isn't enough free space, then cache_step_file() is called
 *    repeatedly until there is enough room for the cached event. The function
 *    returns TRUE on success.  Note that a FALSE return value does not mean
 *    the cache file hasn't been modified.  One or more calls to 
 *    cache_step_file() may have succeeded before the error that killed the
 *    attempt to append the event to the cache file list.  For efficiency
 *    reasons the event is not actually written to disk because multiple
 *    events are buffered before doing a disk write.  If you want to make
 *    sure the disk file is in sync you need to call cache_sync() right after
 *    calling this function.  
 *
 *    The function assumes that the event data includes the terminating
 *    character TECAD_EVENT_END_CHAR.
 */
eif_boolean cache_append_event(tec_event_cache_t *pEC, const char *event)
{
   long eventLen = strlen(event);

   assert(cache_is_open(pEC));

   tecadDEBUG_pf(("cache_append_event(): 0.\n"));
   
   /* Make sure we're not about to corrupt the cache file w/ an event too
    * large to ever get read in again. We have to test against CACHE_RDWR_SZ/2
    * instead of just CACHE_RDWR_SZ otherwise the cache file can be in
    * such a state that cache_dump() won't work. */
   if (eventLen > CACHE_RDWR_SZ/2)
      return FALSE;

   tecadDEBUG_pf(("cache_append_event(): 1.\n"));

   /* Do nothing if the event is too big to fit in the cache file at all. */
   if (eventLen > pEC->maxsz - CACHE_HDR_LEN)
      return TRUE;

   tecadDEBUG_pf(("cache_append_event(): 2.\n"));

   /* Call cache_step_file() until there is enough space for the event. */
   while(eventLen > cache_free_space(pEC))
      if (!cache_step_file(pEC))
         return FALSE;

   tecadDEBUG_pf(("cache_append_event(): 3.\n"));

   /* Allocate write buffer if necessary. */
   if (pEC->wbuf == NULL)
   {
      pEC->tail_to_write = 0;
      pEC->bytes_to_write = 0;
      pEC->wbuf = (char*)malloc(CACHE_RDWR_SZ);
      NULL_CHECK(pEC->wbuf);
   }

   /* Sync with the disk if we don't have enough write buffer space. */
   if (eventLen > CACHE_RDWR_SZ - pEC->bytes_to_write && !cache_sync(pEC))
      return FALSE;

   /* Update cache state in memory. */
   pEC->tail = (pEC->tail == 0 ? CACHE_HDR_LEN : pEC->tail);
   pEC->head = (pEC->head == 0 ? pEC->tail : pEC->head);
   if (pEC->bytes_to_write == 0)
      pEC->tail_to_write = pEC->tail;
   memcpy(pEC->wbuf + pEC->bytes_to_write, event, eventLen);
   pEC->bytes_to_write += eventLen;
   INC_FILE_PTR(pEC->tail, eventLen, pEC->maxsz);

   return TRUE;
}

/* cache_sync()
 *    Syncs the memory state of the event cache with the disk file.  The
 *    function returns TRUE on success and FALSE on failure.  If FALSE is
 *    returned, the list of events stored on disk hasn't been corrupted.
 *    However, the list of "old" events traversed by cache_dump() may
 *    have been changed even if this function returns FALSE.  This is just
 *    a limitation we have to live w/ in this relatively simple implementation.
 *    If you want to make sure the disk file is always in sync with the
 *    memory state of the event cache, just call cache_sync() after every
 *    successful call to cache_step_file() or cache_append_event().
 */
eif_boolean cache_sync(tec_event_cache_t *pEC)
{
   /* First try to write the event data to the file.  Even if this fails
    * somewhere in the middle, the cache file hasn't been corrupted because
    * we haven't updated head and tail yet. */
   if (pEC->bytes_to_write > 0)
   {
      long bytesToWrite, bytesToEnd;

      bytesToWrite = pEC->bytes_to_write;
      bytesToEnd = pEC->maxsz - pEC->tail_to_write;
      if (bytesToWrite < bytesToEnd)
         bytesToEnd = bytesToWrite;
      if (sys_lseek(
             pEC->fd, pEC->tail_to_write, SEEK_SET) !=  pEC->tail_to_write ||
          sys_write(pEC->fd, pEC->wbuf, bytesToEnd) != bytesToEnd)
      {
         tecadDEBUG_pf(("cache_sync(): 1. seek and write failed\n"));
         return FALSE;
      }
      bytesToWrite -= bytesToEnd;
      if (bytesToWrite > 0 &&
          (sys_lseek(pEC->fd, CACHE_HDR_LEN, SEEK_SET) != CACHE_HDR_LEN ||
           sys_write(
              pEC->fd, pEC->wbuf + bytesToEnd, bytesToWrite) != bytesToWrite))
      {
         tecadDEBUG_pf(("cache_sync(): 2. seek and write failed\n"));
         return FALSE;
      }
   }
   
   /* Now update head and tail on disk. */
   {
      char hdr[CACHE_HDR_LEN + 1];
      sprintf(hdr, CACHE_HDR_FORMAT, pEC->maxsz, pEC->head, pEC->tail);
      if (sys_lseek(pEC->fd, 0, SEEK_SET) != 0 ||
          sys_write(pEC->fd, hdr, CACHE_HDR_LEN) != CACHE_HDR_LEN)
      {
         tecadDEBUG_pf(
            ("cache_sync(): 3. seek and write failed\n"));
         return FALSE;
      }
   }

   /* Everything ok. Forget about buffered data to write. */
   pEC->tail_to_write = 0;
   pEC->bytes_to_write = 0;

   return TRUE;
}

/* cache_truncate()
 *    Deletes all the data in the cache file by truncating it to a zero
 *    size file.  The tec_event_cache_t structure is updated to reflect
 *    the change.  The function returns TRUE on success and FALSE on
 *    failure.  Although it depends on the reliablity of the truncation
 *    system call used by sys_truncate(), my hope is that if this function
 *    returns FALSE the cache hasn't been corrupted by a partial 
 *    truncation.
 */
eif_boolean cache_truncate(tec_event_cache_t *pEC)
{
   /* Try truncating the file to a size of zero bytes. */
   if (!sys_truncate(pEC->fd, 0))
      return FALSE;

   /* Update the in memory state. */
   pEC->head = 0;
   pEC->tail = 0;
   pEC->tail_on_open = 0;
   pEC->head_at_read = 0;
   pEC->bytes_read = 0;
   if (pEC->buf != NULL)
   {
      free(pEC->buf);
      pEC->buf = NULL;
   }
   pEC->curr_pos = NULL;
   pEC->tail_to_write = 0;
   pEC->bytes_to_write = 0;
   
   return TRUE;
}

/* cache_dump()
 *    This function can be used to iterate through the contents of the
 *    event cache file.  The callback argument is a pointer to a function
 *    with the signature
 *
 *         eif_boolean callback(char* const  ev,
 *                              eif_boolean  is_queued,
 *                              void        *callback_data)
 *
 *    Starting with the oldest event in the cache, this function iterates
 *    through the events calling callback() for each event.  The ev arg
 *    passed to callback() points to the event data.  The arg is_queued is
 *    passed as TRUE if the event is one awaiting removal from the cache file,
 *    and it's passed as FALSE if it's an old event that has already been
 *    stepped over in the cache.  The callback_data arg is what was given
 *    to cache_dump() in its callback_data arg. If callback returns TRUE,
 *    iteration through the events continues, otherwise cache_dump() cleans up 
 *    and returns.
 *
 *    The behavior is undefined if callback() calls any of the cache_*
 *    functions on the same cache file while doing it's processing.
 *
 *    Cache_dump() returns TRUE if no errors are encountered while trying
 *    to step through the cache file, otherwise FALSE is returned.  Note
 *    that this return value is not related to the return value cache_dump()
 *    gets from the call back function.
 */
eif_boolean cache_dump(const tec_event_cache_t *pEC, 
                       cache_dump_func_t callback,
                       void *callback_data,
                       cache_dump_mode_t dmode)
{
   tec_event_cache_t tEC;
   eif_boolean inQedEvents;
   int ate1stEv;
   eif_boolean retval = FALSE;
   char *ev = NULL;

   assert(cache_is_open(pEC));

   /* Catch special cases where there is nothing to dump. */
   if (pEC->tail == 0 || (dmode == CACHE_DUMP_QUEUED && cache_is_empty(pEC)))
      return TRUE;

   /* Set up temporary event cache structure to iterate through the events. */
   tEC.file_path = pEC->file_path;
   tEC.fd = pEC->fd;
   tEC.maxsz = pEC->maxsz;
   if (dmode == CACHE_DUMP_ALL)
   {
      long fsize;
      if (!sys_get_file_size2(pEC->fd, &fsize))
         return FALSE;
      if (pEC->tail == fsize || pEC->tail == CACHE_HDR_LEN)
         tEC.head = CACHE_HDR_LEN;
      else
         tEC.head = pEC->tail;
   }
   else 
      tEC.head = pEC->head;
   tEC.tail = pEC->tail;
   tEC.tail_on_open = 0;
   tEC.head_at_read = 0;
   tEC.bytes_read = 0;
   tEC.buf = NULL;
   tEC.curr_pos = NULL;
   tEC.curr_step_no = 0;
   tEC.tail_to_write = 0;
   tEC.bytes_to_write = 0;
   tEC.wbuf = NULL;
   tEC.read_only = TRUE;

   /* Iterate through the cache. */
   inQedEvents = (tEC.head == pEC->head);
   ate1stEv = inQedEvents;
   while ((retval = cache_get_curr_event(&tEC, &ev)) && ev != NULL)
   {
      inQedEvents = (inQedEvents || tEC.head == pEC->head);
      if (ate1stEv)
      {
         if (!callback(ev, inQedEvents, callback_data))
            break;
      }
      else
         ate1stEv = TRUE;
      free(ev);
      ev = NULL;
      if (!(retval = cache_step_file(&tEC)))
         break;
   }

   /* Clean up temporary cache struct. */
   if (tEC.buf != NULL) free(tEC.buf);
   return retval;
}

/*****************************************************************************
 **                        Static Helper Functions                          **
 *****************************************************************************/

static char *
filter_events(tec_handle_common_t *ch, char *message, int msg_len)
{
	char    *new_msg;
	char    *nptr;
	char    *mptr = message;
	char    *eptr;
	char    orig_char;
	char    different = 0;

	if (!msg_len)
		return(NULL);

	nptr = new_msg = (char *) malloc((size_t) msg_len + 1);

	if (!new_msg)
		return(NULL);

	/*
	 * Break up 'message' into individual events in case there are more than one
	 */
	while ((eptr = strchr(mptr, TECAD_EVENT_END_CHAR)) != NULL) {
		orig_char = *(++eptr);
		*eptr = '\0';

		if (common_filter_event(ch, mptr, FILTER_EVENT_CACHE_TYPE)==FILTER_KEEP) {
			strcpy(nptr, mptr);
			nptr += (eptr - mptr);
		}
		else
			different = 1;

		*eptr = orig_char;
		mptr = eptr;
	}

	/*
	 * If the new message is not different than the first, return the input event
	 * If the new message is empty, return NULL.
	 * Otherwise, return the new message.
	 */
	if (!different) {
		if (new_msg != NULL)
			free(new_msg);

		new_msg = message;
	}
	else if (nptr == new_msg) {
		if (new_msg != NULL)
			free(new_msg);

		new_msg = NULL;
	}

	return(new_msg);
}

/*****************************************************************************
 **                      Functions Exposed Externally                       **
 *****************************************************************************/

/*
 * tec_buffer_init
 *
 * Purpose:
 *    Initializes some configuration parameters for buffering.
 *
 * Description:
 *
 * Returns:
 *      Nothing
 *
 * Notes/Dependencies:
 *
 */

void
tec_buffer_init(AgentConfig *agent_info)
{
	BufferConfig	*buffer_info = &agent_info->buffer_info;
	char				*env;
	int				megs;
#ifdef WIN32
	static char		exp[MAXPATHLEN];
#endif
#ifdef OS2
	char				*new_fn, *ptr;
#endif
#ifdef OS400
	char		job_name[11];
	char*		strptr;
#endif

	buffer_info->initialized = 1;
	buffer_info->do_buffer = !agent_match(agent_info, "BufferEvents", "NO");

	/*
	 * Event max size in K
	 */
	buffer_info->event_max_size = 4096;

	if ((env = agent_getenv(agent_info, EVENT_MAX_SIZE)) != NULL) {
		if (isdigit(*env))
			buffer_info->event_max_size = atol(env);
	}

	/*
	 * Packet size in bytes
	 */
	buffer_info->maxPacketSize = 0;
	if ((env = agent_getenv(agent_info, MAX_PACKET_SIZE)) != NULL) {
		if (isdigit(*env))
			buffer_info->maxPacketSize = atol(env);
	}

	/*
	 * Max gateway cache file size.
	 * Event adapter buffer file's max size in K
	 */
   if (gw_hack_running_in_gateway)
	{
		megs = 1;
		if ((env = agent_getenv(agent_info, "MaxGWCacheSizeMegs")) != NULL) {
			if (isdigit(*env))
				megs = atoi(env);
		}
      /* We want a larger default buffer size at the TEC gateway. */
      buffer_info->max_size = 1024 * 1024 * megs;
	}
   else
	{
      buffer_info->max_size = 64 * 1024;
	}

	if ((env = agent_getenv(agent_info, "BufEvtMaxSize")) != NULL) {
		if (isdigit(*env))
			buffer_info->max_size = atol(env) * 1024;
	}

	/*
	 * Size, in K, to shrink adapter file's buffer when MaxSize exceeded
    *
    * NOTE - Ignored in the current cache file design.
	 */
	buffer_info->shrink_size = 8 * 1024;

	if ((env = agent_getenv(agent_info, "BufEvtShrinkSize")) != NULL) {
		if (isdigit(*env))
			buffer_info->shrink_size = atol(env) * 1024;
	}
	if (buffer_info->shrink_size < buffer_info->event_max_size)
		buffer_info->shrink_size = buffer_info->event_max_size;

	/*
	 * If BufferFlushRate is not specified, then we will use 0 which means
	 * don't wait at all
	 */
	buffer_info->flush_rate = 0;

	if ((env = agent_getenv(agent_info, "BufferFlushRate")) != NULL) {
		if (isdigit(*env))
			buffer_info->flush_rate = (short) atoi(env);
	}

	/*
	 * Get the size of the buffer file blocks to read.
	 * This MUST be large enough for at least one event.
    *
    * NOTE - Ignored in the current cache file design.
	 */
	buffer_info->read_blk_len = 64 * 1024;

	if ((env = agent_getenv(agent_info, "BufEvtRdBlkLen")) != NULL) {
		if (isdigit(*env))
			buffer_info->read_blk_len = atol(env);
	}

	if ((env = agent_getenv(agent_info, "BufEvtPath")) != NULL) {
		buffer_info->buffer_path = env;

#		ifdef OS2
			/* Change '\' to '/' in filename */
			new_fn = strdup(env);
			NULL_CHECK(new_fn);

			for (ptr = new_fn; *ptr != NULL; ptr++) {
				if (*ptr == '\\')
					*ptr = '/';
			}
			buffer_info->buffer_path = new_fn;
#		endif
	}
	else {
		buffer_info->buffer_path = DEFAULT_BUFFER_PATH;
#ifdef OS400
		/* 
		 * For AS/400, use the job name as the default cache file name
		 */
		memset(job_name, '\0', sizeof(job_name));
		getJobName(job_name);
		/* pull spaces from the job name */
		strptr=strstr(job_name, " ");
		if (strptr != NULL)
			*strptr='\0';
		strcat(buffer_info->buffer_path, job_name);
#endif
	}

#	ifdef WIN32
	if (ExpandEnvironmentStrings(buffer_info->buffer_path, exp, MAXPATHLEN)) {
		buffer_info->buffer_path = strdup(exp);
		NULL_CHECK(buffer_info->buffer_path);
	}
#	endif
}


/*
 * common_buffer_event_to_file
 *
 * Purpose:
 *    Buffers an event or events to a specific file.
 *
 * Description:
 *
 * Returns:
 *      The number of bytes written.
 *
 * Notes/Dependencies:
 *
 */

int
common_buffer_event_to_file_hack(tec_handle_common_t *ch,
                                 char *message,
                                 char *buf_path,
                                 int   ignore_do_buffer)
{
   BufferConfig        *b_info = NULL;
   char                *new_msg = NULL;
   int                  MsgLen;
   tec_event_cache_t    evCacheTmp;
   tec_event_cache_t   *pEC = NULL;
   eif_boolean          cacheAlreadyOpen = FALSE;
   char                *start = NULL, *end = NULL;
   char                 charSave;
   eif_boolean          writeSuccessful;
   int                  nbytes = 0;

#ifndef NO_UTF8
	/*
	 * Strip off the TEC_IPC_UTF8_DELIMITOR if it's present, so that it
	 * won't be written to the buffer file.
	 */
	if ((message != NULL) && (message[0] == TEC_IPC_UTF8_DELIMITOR)) {
		message++;
	}
#endif

	b_info = &ch->agent_config->buffer_info;

	if (!b_info->initialized)
		tec_buffer_init(ch->agent_config);

	/*
	 * Return if BufferEvents was specified, but is not yes or YES
    * WORKAROUND - For the purposes of the TEC gateway I needed to be able
    *    to override b_info->do_buffer and have events be bufferd even
    *    when b_info->do_buffer is FALSE.
	 */
	if (!ignore_do_buffer && !b_info->do_buffer)
		return(-1);

	/*
	 * Check if event should not be cached because it is filtered.
	 * This is different then the 'normal' event filtering and special
	 * ONLY for event that are to be cached. It uses a separate filter
	 * then the normal event filtering and is meant to be more restrictive.
	 * Events may be important enough to send back but not important
	 * enough if they are *first cached* !!
	 */
	MsgLen = strlen(message);
	new_msg = filter_events(ch, message, MsgLen);

	if (new_msg == NULL)
		return(0);
	else if (new_msg != message) {
		message = new_msg;
		MsgLen = strlen(message);
	}
	else 
   {
      /* Making a copy of message because we temporarily modify it below.  A
       * multithreaded caller may not expect that behavior. */
      new_msg = strdup(message);
      NULL_CHECK(new_msg);
   }

	/*
	 * If no buffer path specified, look it up.
	 */
	if (buf_path == NULL)
		buf_path = b_info->buffer_path;

   tecadDEBUG_pf(("common_buffer_event_to_file(): buf_path = %s\n", buf_path));

	/*
	 * Open cache file
	 */
   if (cache_is_open(&ch->event_cache) &&
       strcmp(buf_path, cache_file_name(&ch->event_cache)) == 0)
   {
      cacheAlreadyOpen = TRUE;
      pEC = &ch->event_cache;
   }
   else 
   {
      pEC = &evCacheTmp;
      cache_constructor(pEC);
      if (!cache_open_file(pEC, buf_path, b_info->max_size))
      {
         tecadDEBUG_pf(("common_buffer_event_to_file(): Opening cache file failed.\n"));
         free(new_msg);
         return(-1);         
      }
   }

   /* Try to cache the event(s) to the cache file. */
   nbytes = 0;
   start = new_msg;
   while ((end = strchr(start, TECAD_EVENT_END_CHAR)) != NULL) 
   {
      charSave = end[1];
      end[1] = '\0';
      writeSuccessful = (cache_append_event(pEC, start) && cache_sync(pEC));
      end[1] = charSave;
      if (!writeSuccessful) break;

      nbytes += end - start + 1;
      start = end + 1;
   }

   /* Cleanup */
   if (!cacheAlreadyOpen) cache_destructor(pEC);
   free(new_msg);

   /* BUG: A non-negative return value doesn't indicate that all events
    *      where successfully cached, so an error could have occurred of
    *      which the caller is not made aware.
    */
   return (nbytes == 0 ? -1 : nbytes);
}

int
common_buffer_event_to_file(tec_handle_common_t *ch, char *message,
									 char *buf_path)
{
   return common_buffer_event_to_file_hack(ch, message, buf_path, FALSE);
}

int
common_buffer_event(tec_handle_common_t *ch, char *message)
{
	return(common_buffer_event_to_file_hack(ch, message, NULL, FALSE));
}

void
tec_free_func(void *ptr)
{
	if (ptr)
		free(ptr);
}

/*
//-*****************************************
//  PACKETING CODE SUPPORT
// 
// 
//---------------------------------
*/

/*
flushRate
stepCache
flush_one_at_a_time
flush_many_at_a_time
clearPacket
addEventToPacket
makePacket
submitPacket
*/

/*
 * a data structure to maintain a packet
 */
typedef struct{
  unsigned long  maxSize;
  unsigned long  byteLen;
  int numEvents;
  char *payLoad;
} tec_event_packet_t;

static
int flushRate( int evPerPeriod, int evThisPeriod, int periodLen);

static 
int stepCache(tec_event_cache_t *pEC,int currStep);

static
int flush_one_at_a_time(
  tec_event_cache_t    *pEC,
  tec_handle_common_t *handle,
  int                   evPerPeriod,
  int                   periodLen
);

static
int addEventToPacket (tec_event_packet_t *pkt,char *event);

static 
int makePacket(
  tec_event_packet_t *pkt,
  tec_event_cache_t    *pEC,
  tec_handle_common_t *handle,
  unsigned long                maxPacketLen,
  unsigned long                max_size_packet,
  char **bigEvent
);

static
int submitPacket(
  tec_handle_common_t *handle,
  tec_event_packet_t *pkt,
  char *bigEvent,
  unsigned long *packetCount,
  unsigned long *eventCount
);

static
int flush_many_at_a_time(
  tec_event_cache_t    *pEC,
  tec_handle_common_t *handle,
  int                   evPerPeriod,
  int                   periodLen,
  unsigned long         max_size_packet,
  unsigned long         maxPacketLen
);

static
int hexDump(unsigned char *data,int len,int width)
{
 int j,k,idex;
 unsigned char ch;
 len = ((!len) ? strlen((char *)data) : len);
 idex=0;
 
 printf ("Dumping [%ld] bytes.\n",(long)len);
 printf ("      ");
 for (j=0;j<width;j++) printf("%02d ",j+1);
 printf ("\n");
 while (idex < len)
 {
   printf ("%05d ",idex);
   for (j=0;j<width;j++)
   {
     if (idex+j==len) break; 
     printf("%2x ",data[idex+j]);
   }
   for (k=j;k<width;k++) printf("   "); printf (" | ");

   for (j=0;j<width;j++)
   {
     if (idex+j==len) break; 
     ch =data[idex+j];
     if ((int)ch > 126  || (int)ch < 33)
       printf("%c",'.');
     else
       printf("%c", (char )ch);
   }
   printf("\n");
   idex+=j;
 }
 return 0;
}
static
int hexDumpToStr(char *result,unsigned char *data,int len,int width)
{
 int j,k,idex;
 unsigned char ch;
 char obuff[1024];
 len = ((!len) ? strlen((char *)data) : len);
 idex=0;
 
 strcpy(result,"");
 sprintf (obuff,"Dumping [%ld] bytes.\n",(long)len);
 strcat(result,obuff);
 sprintf (obuff,"      ");
 strcat(result,obuff);
 for (j=0;j<width;j++){
   sprintf(obuff,"%02d ",j+1);
   strcat(result,obuff);
 }
 sprintf (obuff,"\n");
 strcat(result,obuff);
 while (idex < len)
 {
   sprintf (obuff,"%05d ",idex);
   strcat(result,obuff);
   for (j=0;j<width;j++)
   {
     if (idex+j==len) break; 
     sprintf(obuff,"%2x ",data[idex+j]);
     strcat(result,obuff);
   }
   for (k=j;k<width;k++){
      sprintf(obuff,"   "); 
      strcat(result,obuff);
   }
   sprintf (obuff," | ");
   strcat(result,obuff);

   for (j=0;j<width;j++)
   {
     if (idex+j==len) break; 
     ch =data[idex+j];
     if ((int)ch > 126  || (int)ch < 33)
     {
       sprintf(obuff,"%c",'.');
       strcat(result,obuff);
     }
     else
     {
       sprintf(obuff,"%c", (char )ch);
       strcat(result,obuff);
     }
   }
   sprintf(obuff,"\n");
   strcat(result,obuff);
   idex+=j;
 }
 return 0;
}
static
int hexDumpIt(unsigned char *data,int len)
{
 return( hexDump(data,len,20));
}
static
int hexDumper(unsigned char *data)
{
 return( hexDump(data,0,20));
}

static
int clearPacket(tec_event_packet_t *pkt)
{
 memset (pkt->payLoad,'\0',pkt->maxSize);
 pkt->byteLen=0;
 pkt->numEvents=0;
 return 0;
}

/*---------------------------------*/
static
int flushRate(
   int                   evPerPeriod,
   int                   evThisPeriod,
   int                   periodLen
)
{
      /* Deal with flush rate. */
      if (evPerPeriod > 0)
      {
         /*++evThisPeriod;*/
         if (evThisPeriod >= evPerPeriod) {
#ifndef OS400
            sleep(periodLen);
#else
            usleep(periodLen);
#endif
            evThisPeriod = 0;
         }
      }
  return evThisPeriod;
}


static 
int stepCache(tec_event_cache_t *pEC,int currStep)
{
  	   /* Step forward in the cache only if the call to tec_put_event()
  	     * didn't end up making a call to cache_step_file() in order to
  	     * make room for a recaching of the event at the end of the event
  	     * list. */
  	   if ( 
              (currStep == cache_curr_step_no(pEC))
           && !(cache_step_file(pEC))
           && cache_sync(pEC)
  	      )
  	    {
         return 1;
       }
       return 0;
}

/* ----------------------------------*/
static
int flush_one_at_a_time(
   tec_event_cache_t    *pEC,
   tec_handle_common_t *handle,
   int                  evPerPeriod,
   int                  periodLen
)
{
   int    evThisPeriod=0;
   int    curr_step_no;
   char  *event;
   int    sent = 0;

   tecadDEBUG_pf(("flush_one_at_a_time(): \n"));
   curr_step_no = cache_curr_step_no(pEC);
   while (
              !cache_done_with_ev_cached_at_open(pEC) 
           &&  cache_get_curr_event(pEC, &event)
         )
   {
      /* If cache_done_with_ev_cached_at_open() works correctly, then event
       * can't be NULL. */
      assert(event != NULL);

      if (tec_put_event(handle, event) > 0)
         ++sent;
      free(event);
     
      if (stepCache(pEC,curr_step_no))
  	   {
         return -3; /*  goto  EXIT; */
      }
      curr_step_no = cache_curr_step_no(pEC);

      if (evPerPeriod > 0){
        evThisPeriod++;
        evThisPeriod= flushRate(evPerPeriod,evThisPeriod,periodLen);
      } 
   }
   return sent;
}

/*------------------*/
static
int addEventToPacket (tec_event_packet_t *pkt, char *event)
{
   if (debugg){
     printf("add [%s]\n",event);
     /*hexDump(event,0,10);*/
   }
   strcat(pkt->payLoad,event);
   pkt->byteLen+=strlen(event);
   pkt->numEvents++;
   return pkt->byteLen;
}

/*
//------------------
// continue to concatenate events to
// the packet until threshhold
// or cache is empty
//
// return packetLen - number of events
//
*/
static 
int makePacket(
  tec_event_packet_t  *pkt,
  tec_event_cache_t   *pEC,
  tec_handle_common_t *handle,
  unsigned long                maxPacketLen,
  unsigned long                max_size_packet,
  char **bigEvent
)
{
   char * event=NULL;
   int       curr_step_no=0;
   int       eventLen    =0;
   int       bufferReady =0;
   int rv=0;

   tecadDEBUG_pf(("makePacket(): \n"));
   clearPacket(pkt);

   curr_step_no = cache_curr_step_no(pEC);

   if (debugg)
     printf("\nmaking packet...\n");

   while (
              (!bufferReady)
           && (!cache_done_with_ev_cached_at_open(pEC) )
           && (cache_get_curr_event(pEC, &event))
         )
   {
     /* If cache_done_with_ev_cached_at_open() works correctly, then event can't be NULL. */
     assert(event != NULL);
     eventLen=strlen(event);

     /*
     // buffer ready when
     //    maximum size of packet reached
     // or number sent is has reached maxium 
     */

     bufferReady=(   ((pkt->byteLen+eventLen+1) >= max_size_packet) 
                  || ( maxPacketLen ? (pkt->numEvents >= maxPacketLen) : 0)
                 );

     if (!bufferReady) 
	  {
      /*
	    // add event to packet
       // and step the file
       */
  	   addEventToPacket (pkt,event);
      tecadDEBUG_pf(("makePacket(): add\n"));

      if (stepCache(pEC,curr_step_no))
  	   {
         /* problem stepping */
         rv=-1;
  	      break;
  	   }
  	   curr_step_no = cache_curr_step_no(pEC);
     }
     else
     { /* buffer ready */
        tecadDEBUG_pf(("makePacket(): bufferReady\n"));

  	     if (!pkt->numEvents )
  	     {
          /*
  	       // Event bigger than than the a packet
  	       // printf("BIG EVENT send it\n");
          */

  	       pkt->numEvents=1;
          if (stepCache(pEC,curr_step_no))
  	       {
            /* problem stepping */
            rv=-1;
  	       }
  	       else
  	         *bigEvent=event;
          break;
  	     }
      
     } /* else bufferready */

     free(event);
     event=NULL;

   }  /* while */
   if (event!=NULL)
     if (bigEvent==NULL)
  	   free(event);

   rv= ((!rv) ? pkt->numEvents : rv);
       
   tecadDEBUG_pf(("makePacket(): returning %i\n",rv));
   if (debugg)
     printf ("rv %i\n",rv);
   return rv;
}

/*===========================*/
static
int submitPacket(
	tec_handle_common_t *handle,
	tec_event_packet_t *pkt,
	char *bigEvent,
	unsigned long *packetCount,
	unsigned long *eventCount
)
{
   int bytesSent=0;

#ifdef DEBUG
   char *dumpPacket=NULL;
   unsigned long MaxSizeDumpStr=40000;
#endif

#ifdef DEBUG
   dumpPacket=(char *)malloc(MaxSizeDumpStr);
#endif
   tecadDEBUG_pf(("submitPacket(): \n"));
   if (debugg)
     printf ("Put packet\n");
   /* Send the packet */
   if (bigEvent!=NULL)
   {
      if (debugg)
      {
        printf ("Put bigEvent\n");
        hexDumper(bigEvent);
      }

#ifdef DEBUG
      hexDumpToStr(dumpPacket, bigEvent,0,20);
      tecadDEBUG_pf(("submitPacket(): bigEvent\n%s\n",dumpPacket));
#endif

      bytesSent=tec_put_event(handle, bigEvent);
   }
   else
   {
      if (debugg)
      {
        printf ("Put packet payload\n");
        hexDumpIt( pkt->payLoad,(int)pkt->byteLen+2);
      }
#ifdef DEBUG
      hexDumpToStr(dumpPacket, pkt->payLoad,(int)pkt->byteLen+2,20);
      tecadDEBUG_pf(("submitPacket(): payload\n%s\n",dumpPacket));
#endif

      bytesSent=tec_put_event(handle, pkt->payLoad);
   }
   *packetCount+=1;

	if (bytesSent > 0)
	{ 
       *eventCount+=pkt->numEvents;
       if (debugg)
         printf("Sent [%ld] bytes of [%ld] bytes.\n",(long)bytesSent,(unsigned long)pkt->byteLen);
	}
   else
   {
       if (debugg)
       printf ("Nothing sent [%i]\n",bytesSent);
   }

#ifdef DEBUG
   free(dumpPacket);
#endif
   return bytesSent;
}

/*
//-**************************************
//  -2   error malloc
//  -1   error
//
*/
static
int flush_many_at_a_time(
  tec_event_cache_t    *pEC,
  tec_handle_common_t *handle,
  int                   evPerPeriod,
  int                   periodLen,
  unsigned long         max_size_packet,
  unsigned long         maxPacketLen
)
{
   tec_event_packet_t  pkt;  /* a single packet to be used over */
   char		*bigEvent=NULL;  /* pointer to a big event */
   int		error_malloc=0;  /* error in mem allocation */

   int evThisPeriod  =0;   /* events occurred during this period */
   int bytesSent     =0;   /* bytesSent on a single submit */
   unsigned long packetCount   =0;   /* count of packets created and sent */
   unsigned long eventCount    =0;   /* count of events sent in packets */
   int makingPackets =1;   /* loop control */

   int rv=0;  /* return value */
   int rc=0;  /* return code */

   tecadDEBUG_pf(("flush_many_at_a_time():\n"));

   if (debugg)
     printf("flushing many...\n");

    /* Allocate memory for the events buffer that munst contain multiple events*/ 
   pkt.maxSize=max_size_packet;
   pkt.payLoad=(char*)malloc(pkt.maxSize);
	if (!pkt.payLoad) 
   {
		error_malloc=1;
		return -2;
   }
   clearPacket(&pkt);

   maxPacketLen = ((evPerPeriod > 0) ? evPerPeriod : 0);

   tecadDEBUG_pf(("flush_many_at_a_time(): maxPacketLen=[%ld]\n",maxPacketLen));
   if (debugg)
     printf("making packets...\n");

   while (makingPackets)
   {
       rc=makePacket(&pkt,pEC,handle,maxPacketLen,max_size_packet,&bigEvent);

       tecadDEBUG_pf(("flush_many_at_a_time(): makePacket rc=[%d]\n",rc));
       evThisPeriod+=((pkt.numEvents > 0) ? pkt.numEvents : 0);
       if (pkt.numEvents > 0) 
	    {
          bytesSent= submitPacket(handle,&pkt,bigEvent,&packetCount,&eventCount);
          clearPacket(&pkt);
       } 
       else /* packetlen <=0 */
       {
         tecadDEBUG_pf(("flush_many_at_a_time(): empty packet\n"));
         if (debugg)
            printf("empty packet.\n");

         makingPackets=0;
         if (rc <0)
            rv=rc;
       } 
       if (bigEvent!=NULL)
       {
          tecadDEBUG_pf(("flush_many_at_a_time(): bigEvent being freed\n"));
          free(bigEvent);
          bigEvent=NULL;
       }
       if (evPerPeriod > 0)
        evThisPeriod= flushRate(evPerPeriod,evThisPeriod,periodLen);

   }  /* end while making packets */

   if (!rv)
     rv=eventCount;
   clearPacket(&pkt);
   free(pkt.payLoad);
   

   if (debugg)
   {
     printf("[%ld] packets generated.\n",packetCount);
     printf("[%ld] events.\n",eventCount);
   }
   tecadDEBUG_pf(("flush_many_at_a_time(): [%ld] packets generated\n",packetCount));
   tecadDEBUG_pf(("flush_many_at_a_time(): [%ld] events processed\n",eventCount));

   return rv;
}

/*
 * Flush all events in the event buffer cache (on disk).
 * Try to send each event. If it won't go, then it will
 * will get re-buffered again.  Return the number of events
 * we tried to send.
 */

int
tec_flush_events_from_file(tec_handle_t handle, char *buf_path)
{
	BufferConfig			*b_info;
	tec_handle_common_t	*ch = (tec_handle_common_t *) handle;
   tec_event_cache_t    *pEC = &ch->event_cache;
   int                   total_sent = 0;
   int                   evPerPeriod=0;
   int                   periodLen=0;
   unsigned long		    max_size_packet=8192;
   long		             maxPacketLen=1;
   int					    error_malloc=0;
   int multiSend=0;

   if (debugg)
     printf("tec_flush_events_from_file\n");
   /* Are we trying to reenter tec_flush_events_from_file()? */
	assert(!ch->in_flush_events);

   /* Is somebody else messing w/ the event cache? */
   assert(!cache_is_open(pEC));

   /* Set in_flush_events, and _don't_ return w/out clearing it. */
	ch->in_flush_events = 1;

	b_info = &ch->agent_config->buffer_info;

	if (!b_info->initialized)
		tec_buffer_init(ch->agent_config);

	/*
	 * If no buffer path specified, look it up.
	 */
	if (buf_path == NULL)
		buf_path = b_info->buffer_path;

   if (!cache_open_file(pEC, buf_path, b_info->max_size))
      goto EXIT;

   if (b_info->maxPacketSize)
   {
     multiSend=1;
     max_size_packet= b_info->maxPacketSize;
   }
   maxPacketLen=0;


	/* It's not necessary to filter because the cached events are already filterd by tec_put_event */  
/*//	ch->dont_filter=1; */

   if (b_info->flush_rate > 0)
   {
      /* The exact rate may be somewhat different than what the user
       * specified because the sleep function we're using has a granularity
       * of 1 sec.  We could be more exact even w/ this 1 second granularity
       * by varying evPerPeriod and periodLen over time, but I think that
       * would be hard. */
      if (b_info->flush_rate >= 60)
      {
         evPerPeriod = b_info->flush_rate/60;
         if (((10*b_info->flush_rate)/60) % 10 >= 5)
            ++evPerPeriod;
         periodLen = 1 /* sec */;
      }
      else
      {
         evPerPeriod = 1;
         periodLen = 60/b_info->flush_rate;
         if ((600/b_info->flush_rate) % 10 >= 5)
            ++periodLen;
      }
   }
   else
      evPerPeriod = 0;

   if (multiSend)
		total_sent=flush_many_at_a_time(
		  (tec_event_cache_t    *)pEC,
		  (tec_handle_common_t  *)handle,
		  (int)                   evPerPeriod,
		  (int)                   periodLen,
		  (unsigned long)                 max_size_packet,
		  (unsigned long)                 maxPacketLen
		);
	else
		total_sent=flush_one_at_a_time(
	     (tec_event_cache_t    *)pEC,
  		  (tec_handle_common_t  *)handle,
		  (int)                   evPerPeriod,
		  (int)                   periodLen
     );

   if (total_sent>0)
   {
     if (cache_is_empty(pEC))
      cache_truncate(pEC);
     else if (cache_max_size(pEC) != (long)b_info->max_size)
      /* Resize the cache file if a new maximum size has been specified. */
      cache_resize_file(pEC, b_info->max_size);
  }

  EXIT:
   /* BUG: Since the return value is the count of events processed from
    *      the cache file, the caller really has no idea of whether or
    *      not things went 100% successfully.
    */
   if (!error_malloc)
   {
      /*
		// tec_put_event(handle, events);
		// total_sent=total_sent+sent;
		//free(events);
      */
   }	
   cache_destructor(pEC);
   ch->in_flush_events = 0;
	return(total_sent);
}

/* This version of tec_flush_events_from_file calls a callback function on
 * each flushed event rather than calling tec_put_event().  This version
 * also doesn't alter or test handle->in_flush_events.
 * Returns: 
 *    0 - All events flushed successfully.
 *    1 - All events not flushed successfully because of some I/O error 
 *        associated w/ the cache file.
 *    2 - All events not flushed successfully because the callback returned
 *        FALSE for some event.  The event on which FALSE was returned 
 *        remains in the cache file.
 */
int
tec_flush_events_from_file2(tec_handle_t handle,
                            char *buf_path,
                            FlushCallback cbFunc,
                            void *callback_data)
{
	BufferConfig			*b_info;
	tec_handle_common_t	*ch = (tec_handle_common_t *) handle;
   tec_event_cache_t    *pEC = &ch->event_cache;
   char                 *event;
   int                   curr_step_no;
   int                   evPerPeriod;
   int                   evThisPeriod;
   int                   periodLen=0;
   int                   rc = 0;
   int                   cbResult = 0;
#ifdef OS400
   char	                command[300];
#endif
   /* Is somebody else messing w/ the event cache? */
   assert(!cache_is_open(pEC));

	b_info = &ch->agent_config->buffer_info;

	if (!b_info->initialized)
		tec_buffer_init(ch->agent_config);

	/*
	 * If no buffer path specified, look it up.
	 */
	if (buf_path == NULL)
		buf_path = b_info->buffer_path;

   if (!cache_open_file(pEC, buf_path, b_info->max_size))
   {
      rc = 1;
      goto EXIT;
   }

   if (b_info->flush_rate > 0)
   {
      /* The exact rate may be somewhat different than what the user
       * specified because the sleep function we're using has a granularity
       * of 1 sec.  We could be more exact even w/ this 1 second granularity
       * by varying evPerPeriod and periodLen over time, but I think that
       * would be hard. */
      if (b_info->flush_rate >= 60)
      {
         evPerPeriod = b_info->flush_rate/60;
         if (((10*b_info->flush_rate)/60) % 10 >= 5)
            ++evPerPeriod;
         periodLen = 1 /* sec */;
      }
      else
      {
         evPerPeriod = 1;
         periodLen = 60/b_info->flush_rate;
         if ((600/b_info->flush_rate) % 10 >= 5)
            ++periodLen;
      }
   }
   else
      evPerPeriod = 0;
   evThisPeriod = 0;
   curr_step_no = cache_curr_step_no(pEC);
   while (!cache_done_with_ev_cached_at_open(pEC) &&
          cache_get_curr_event(pEC, &event))
   {
      /* If cache_done_with_ev_cached_at_open() works correctly, then event
       * can't be NULL. */
      assert(event != NULL);
      
      cbResult = cbFunc(event, callback_data);
      free(event);
      if (!cbResult) 
      {
         rc = 2;
         goto EXIT;
      }

				  /* Step forward in the cache only if the call to the callback function
					* didn't end up making a call to cache_step_file(). */
				  if (curr_step_no == cache_curr_step_no(pEC) &&
						!(cache_step_file(pEC) && cache_sync(pEC)))
					  goto EXIT;

				  curr_step_no = cache_curr_step_no(pEC);

				  /* Deal with flush rate. */
				  if (evPerPeriod > 0)
				  {
					  ++evThisPeriod;
					  if (evThisPeriod >= evPerPeriod) {
		  #ifndef OS400
						  sleep(periodLen);
		  #else
					usleep(periodLen);
		  #endif
						  evThisPeriod = 0;
					  }
				  }
			  }

			  if (cache_is_empty(pEC))
				  cache_truncate(pEC);

			 EXIT:
			  cache_destructor(pEC);
			  return rc;
		  }

		  long
		  tec_event_buffer_size(tec_handle_t in_ch, char *buf_path,HANDLE *fd_ptr)
		  {
			  tec_handle_common_t  *ch = (tec_handle_common_t*)in_ch;
			  BufferConfig			*b_info = NULL;
			  HANDLE					BufEvtFD;
			  long                 bufferSize;
		  #ifdef OS400
			  char                 command[300];
#endif

	if (fd_ptr != NULL)
		*fd_ptr = INVALID_HANDLE_VALUE;

	if ((ch != NULL) && (ch->agent_config != NULL))
   {
      b_info = &ch->agent_config->buffer_info;
      if (!b_info->initialized)
         tec_buffer_init(ch->agent_config);
   }

	/*
	 * If no buffer path specified, look it up.
	 */
	if (buf_path == NULL)
   {
      if (b_info == NULL) return -1;
		buf_path = b_info->buffer_path;
   }
      /*************************************************************/
      /* Just open & truncate the file instead of using O_TRUNC    */
      /* option on the OPEN. ftruncate discards the file contents. */
      /*************************************************************/
#ifdef OS400
   if ((BufEvtFD=open(buf_path, (O_RDWR | O_APPEND), 0 )) == -1) { /* @A1 A*/
      if ( errno == EBUSY )                                      /* @A1 A */
      {                                                          /* @A1 A */
          sendPgmMsg(TEC_CANNOT_BUF, TEC_ADP_MSG_FILE_Q,         /* @A1 A */
                     NULL, '\0', "*DIAG     " );                 /* @A1 A */
          return(0);                                             /* @A1 A */
      }                                                          /* @A1 A */
      else {
      /* Create the file if it does not exist and change owner to QSYS @A1 A*/
        if ((BufEvtFD=open(buf_path,
                          (O_RDWR | O_APPEND | O_CREAT),
                          (S_IRWXU | S_IRGRP | S_IROTH))) == -1) { /* @A1 C */
             sendPgmMsg(TEC_CANNOT_BUF, TEC_ADP_MSG_FILE_Q,      /* @A1 A */
                        NULL, '\0', "*DIAG     " );              /* @A1 A */
             return(0);                                          /* @A1 C */
         }

         sprintf(command, "QSYS/CHGOWN OBJ('%s') NEWOWN(QSYS)",  /* @A1 A */
                 buf_path);                                      /* @A1 A */
         QCMDEXC (command, strlen(command));                     /* @A1 A */
      }                                                          /* @A1 A */
   }                                                             /* @A1 A */
#else
   if ((BufEvtFD = sys_open_file(buf_path, SYS_O_RDWR)) == INVALID_HANDLE_VALUE)
      return -1;
#endif
   if (!sys_get_file_size2(BufEvtFD, &bufferSize))
   {
      sys_close_file(BufEvtFD);
      return -1;
   }

	if (fd_ptr != NULL)
		*fd_ptr = BufEvtFD;
	else
		sys_close_file(BufEvtFD);

   return bufferSize;
}
